from functools import wraps as _wraps

import numpy as _np
from gustaf import Edges as _Edges
from gustaf import Faces as _Faces
from gustaf import Vertices as _Vertices
from gustaf import Volumes as _Volumes
from gustaf.create.edges import from_data as _from_data
from gustaf.utils import connec as _connec
from gustaf.utils.arr import enforce_len as _enforce_len

from splinepy import settings as _settings
from splinepy.helpme import visualize as _visualize
from splinepy.utils import log as _log
from splinepy.utils.data import cartesian_product as _cartesian_product


def edges(
    spline,
    resolution=100,
    all_knots=True,
):
    """Extract edges (lines) from a given spline, or multipatch object. Only
    entity you can extract without dimension limit.

    Parameters
    -----------
    spline: Spline/ Multipatch
      (`self`-argument if called via extract member of a spline)
    resolution: int or list
      samples per parametric dimension
    all_knots: bool
      Switch to allow all knot-line extraction or just contour (para_dim>0)

    Returns
    --------
    edges: Edges
    """
    from splinepy import Multipatch as _Multipatch

    if spline.para_dim == 1:
        # get a single value
        if isinstance(resolution, (_np.ndarray, tuple, list)):
            resolution = max(resolution)
        # ensure int
        resolution = int(resolution)

        vertices = spline.sample(resolution)
        e = _connec.range_to_edges(
            (0, resolution),
            closed=False,
        )
        if isinstance(spline, _Multipatch):
            e = _np.vstack(
                [e + i * resolution for i in range(len(spline.patches))]
            )
        return _Edges(
            vertices=vertices,
            edges=e,
        )

    else:
        # for para_dim > 1, we support array-like resolutions
        resolution = _enforce_len(resolution, spline.para_dim)

        # recursion for Multipatch splines
        if isinstance(spline, _Multipatch):
            return _Edges.concat(
                [
                    edges(spline=s, resolution=resolution, all_knots=all_knots)
                    for s in spline.patches
                ]
            )

        # Single patch case
        if all_knots:
            relevant_knots = spline.unique_knots
        else:
            relevant_knots = list(spline.parametric_bounds.T)

        temp_edges = []  # edges' is not a valid syntax
        for i in range(spline.para_dim):
            split_knots = relevant_knots.copy()
            split_knots.pop(i)
            n_knot_lines = _np.prod([len(s) for s in split_knots], dtype=int)
            # Create query points
            # Sampling along a line can be done using cartesian product,
            # however, to preserve the correct order we need to permute the
            # columns, example:
            #
            # [y0, x, z]    [x, y0, z]
            # [y1, x, z] -> [x, y1, z]
            # [y2, x, z]    [x, y2, z]
            reorder_mask = (
                [*range(1, i + 1)] + [0] + [*range(1 + i, spline.para_dim)]
            )
            extract_knot_queries = _cartesian_product(
                [_np.linspace(*spline.parametric_bounds[:, i], resolution[i])]
                + split_knots,
                reverse=True,
            )[:, reorder_mask]

            # Retrieve connectivity to be repeated
            single_line_connectivity = _connec.range_to_edges(
                (0, resolution[i]),
                closed=False,
            )

            temp_edges.append(
                _Edges(
                    vertices=spline.evaluate(extract_knot_queries),
                    edges=_np.vstack(
                        [
                            single_line_connectivity + resolution[i] * j
                            for j in range(n_knot_lines)
                        ]
                    ),
                )
            )

        return _Edges.concat(temp_edges)


def _uniform_3d_faces(spline, res):
    # Extract boundaries and sample from boundaries
    if isinstance(spline, _settings.NAME_TO_TYPE["Multipatch"]):
        boundaries = spline.boundary_multipatch()
    else:
        boundaries = _settings.NAME_TO_TYPE["Multipatch"](
            splines=spline.extract.boundaries()
        )

    n_faces = (res - 1) ** 2
    vertices = res**2
    f_loc = _connec.make_quad_faces(_enforce_len(res, 2))
    face_connectivity = _np.empty((n_faces * len(boundaries.patches), 4))

    # Create Connectivity for Multipatches
    for i in range(len(boundaries.patches)):
        face_connectivity[i * n_faces : (i + 1) * n_faces] = f_loc + (
            i * vertices
        )

    # make faces and merge vertices before returning
    return _Faces(vertices=boundaries.sample(res), faces=face_connectivity)


def faces(
    spline,
    resolution,
    watertight=True,
):
    """Extract faces from spline. Valid iff para_dim is one of the following:
    {2, 3}. In case of {3}, it will return only surfaces. If internal faces are
    desired, used `spline.extract.volumes().faces()`. Note that dimension
    higher than 3 is not showable.

    Parameters
    -----------
    spline: Spline / Multipatch
      (`self`-argument if called via extract member of a spline)
    resolution: int or list
      samples per parametric dimension.
    watertight: bool
      Default is True. Only related to para_dim = 3 splines. If False,
      overlapping vertices at boundary edges won't be merged.

    Returns
    --------
    faces: faces
    """
    from splinepy import Multipatch as _Multipatch

    if spline.para_dim == 2:
        if isinstance(spline, _Multipatch):
            n_faces = (resolution - 1) ** spline.para_dim
            vertices = resolution**spline.para_dim
            f_loc = _connec.make_quad_faces(
                _enforce_len(resolution, spline.para_dim)
            )
            face_connectivity = _np.empty((n_faces * len(spline.patches), 4))
            # Create Connectivity for Multipatches
            for i in range(len(spline.patches)):
                face_connectivity[i * n_faces : (i + 1) * n_faces] = f_loc + (
                    i * vertices
                )
        else:
            face_connectivity = _connec.make_quad_faces(
                _enforce_len(resolution, spline.para_dim)
            )
        faces = _Faces(
            vertices=spline.sample(resolution),
            faces=face_connectivity,
        )

    elif spline.para_dim == 3:
        full_res = _enforce_len(resolution, spline.para_dim)
        if (full_res[0] == full_res[1] == full_res[2]) or isinstance(
            spline, _Multipatch
        ):
            faces = _uniform_3d_faces(spline, int(max(full_res)))
        else:
            # single patch should support varying resolution sample
            b_faces = []
            for i, bs in enumerate(spline.extract.boundaries()):
                res = full_res.tolist()
                pop_dim = i // 2
                res.pop(pop_dim)

                b_faces.append(
                    _Faces(bs.sample(res), _connec.make_quad_faces(res))
                )

            faces = _Faces.concat(b_faces)

    else:
        raise ValueError("Invalid spline to make faces.")

    if watertight:
        faces.merge_vertices()

    return faces


def volumes(spline, resolution, watertight=False):
    """Extract volumes from spline. Valid iff spline.para_dim == 3.

    Parameters
    -----------
    spline: Spline / Multipatch
      (`self`-argument if called via extract member of a spline)
    resolution: int or list
      Sample resolution
    watertight : bool
      Return watertight mesh

    Returns
    --------
    volumes: Volumes
    """
    from splinepy import Multipatch as _Multipatch

    if spline.para_dim != 3 or spline.dim != 3:
        raise ValueError(
            "Volume extraction from a spline is only valid for "
            "para_dim: 3 dim: 3 splines."
        )

    if isinstance(spline, _Multipatch):
        if not isinstance(resolution, int):
            raise ValueError(
                "Resolution for sampling must be integer type for Multipatch "
                "objects"
            )
        p_connect = _connec.make_hexa_volumes(_enforce_len(resolution, 3))
        n_elements_per_patch = p_connect.shape[0]
        n_vertices_per_patch = resolution**spline.para_dim
        connectivity = _np.empty(
            (n_elements_per_patch * len(spline.patches), p_connect.shape[1])
        )

        for i in range(len(spline.patches)):
            ids = slice(
                i * n_elements_per_patch,
                (i + 1) * n_elements_per_patch,
                None,
            )
            connectivity[ids] = p_connect + i * n_vertices_per_patch
    else:
        connectivity = _connec.make_hexa_volumes(_enforce_len(resolution, 3))

    # Create volumes
    volumes = _Volumes(
        vertices=spline.sample(resolution),
        volumes=connectivity,
    )

    if watertight:
        volumes.merge_vertices()

    return volumes


def control_points(spline):
    """Extracts control points and return as vertices. Same can be achieved by
    doing `gustaf.Vertices(spline.control_points)`

    Parameters
    -----------
    spline: Spline / Multipatch
      (`self`-argument if called via extract member of a spline)

    Returns
    --------
    cps_as_Vertices: Vertices
    """
    return _Vertices(spline.control_points)


def control_edges(spline):
    """Extract control edges (mesh). Valid iff para_dim is 1.

    Parameters
    -----------
    spline: Spline / Multipatch
      (`self`-argument if called via extract member of a spline)

    Returns
    --------
    edges: Edges
    """
    from splinepy import Multipatch as _Multipatch

    if spline.para_dim != 1:
        raise ValueError("Invalid spline type!")

    if isinstance(spline, _Multipatch):
        # @todo avoid loop and transfer range_to_edges to cpp
        return _Edges.concat([control_edges(s) for s in spline.patches])
    else:
        return _Edges(
            vertices=spline.control_points,
            edges=_connec.range_to_edges(
                len(spline.control_points), closed=False
            ),
        )


def control_faces(spline):
    """Extract control face (mesh). Valid iff para_dim is 2.

    Parameters
    -----------
    spline: Spline / Multipatch
      (`self`-argument if called via extract member of a spline)

    Returns
    --------
    faces: Faces
    """
    from splinepy import Multipatch as _Multipatch

    if spline.para_dim != 2:
        raise ValueError("Invalid spline type!")

    if isinstance(spline, _Multipatch):
        # @todo avoid loop and transfer range_to_edges to cpp
        return _Faces.concat([control_faces(s) for s in spline.patches])
    else:
        return _Faces(
            vertices=spline.control_points,
            faces=_connec.make_quad_faces(spline.control_mesh_resolutions),
        )


def control_volumes(spline):
    """Extract control volumes (mesh). Valid iff para_dim is 3.

    Parameters
    -----------
    spline: Spline / Multipatch
      (`self`-argument if called via extract member of a spline)
    Returns
    --------
    volumes: Volumes
    """
    from splinepy import Multipatch as _Multipatch

    if spline.para_dim != 3:
        raise ValueError("Invalid spline type!")

    if isinstance(spline, _Multipatch):
        # @todo avoid loop and transfer range_to_edges to cpp
        return _Volumes.concat([control_volumes(s) for s in spline.patches])
    else:
        return _Volumes(
            vertices=spline.control_points,
            volumes=_connec.make_hexa_volumes(spline.control_mesh_resolutions),
        )


def control_mesh(spline):
    """Calls control_edges, control_faces, control_volumes based on current
    spline.

    Parameters
    -----------
    spline: Spline / Multipatch
      (`self`-argument if called via extract member of a spline)

    Returns
    --------
    control_mesh: Edges or Faces or Volumes
    """
    if spline.para_dim == 1:
        return control_edges(spline)
    elif spline.para_dim == 2:
        return control_faces(spline)
    elif spline.para_dim == 3:
        return control_volumes(spline)
    else:
        raise ValueError(
            "Invalid para_dim to extract control_mesh. Supports 1 to 3."
        )


def spline(spline, para_dim, split_plane):
    """Extract a sub spline from a given representation.

    Parameters
    ----------
    spline: Spline / Multipatch
      (`self`-argument if called via extract member of a spline)
    para_dim : int
      parametric dimension to be extracted
    split_plane : float / tuple<float, float>
      interval or value in parametric space to be extracted from the spline
      representation

    Returns
    -------
    spline
    """
    from splinepy.spline import Spline as _Spline

    # Check type
    if not isinstance(spline, _Spline):
        raise TypeError("Unknown spline representation passed to sub spline")

    # Check arguments for sanity
    if para_dim > spline.para_dim:
        raise ValueError(
            "Requested parametric dimension exceeds spline's parametric"
            " dimensionality."
        )

    # create 1d array for easy type / size check
    split_plane = _np.asanyarray(split_plane).ravel()
    if not _np.issubdtype(split_plane.dtype, _np.floating):
        raise ValueError("split_plane must be floating type.")

    if split_plane.size > 2:
        raise ValueError("split_plane must have size of 1 or 2.")

    is_interval = len(split_plane) == 2

    # Check if is bezier-type
    if not spline.has_knot_vectors:
        spline_copy = spline.nurbs if spline.is_rational else spline.bspline
    else:
        spline_copy = spline.copy()

    for _ in range(spline_copy.degrees[para_dim] + 1):
        # Will do nothing if spline already has sufficient number of knots
        # at given position go to c^-1 extraction
        spline_copy.insert_knots(para_dim, split_plane)

    # start and end id. indices correspond to [first dim][first appearance]
    start_id = (
        _np.where(
            abs(spline_copy.knot_vectors[para_dim].numpy() - split_plane[0])
            < _settings.TOLERANCE
        )[0][-1]
        - spline_copy.degrees[para_dim]
    )

    end_id = (
        _np.where(
            abs(spline_copy.knot_vectors[para_dim].numpy() - split_plane[-1])
            < _settings.TOLERANCE
        )[0][-1]
        - spline_copy.degrees[para_dim]
    )

    # Use MultiIndex for extraction
    mi = spline_copy.multi_index
    indices = [slice(None, None, None) for _ in range(spline_copy.para_dim)]
    if not is_interval:
        # Check if boundary
        if (
            abs(spline_copy.knot_vectors[para_dim][-1] - split_plane[-1])
            < _settings.TOLERANCE
        ):
            start_id -= 1
        indices[para_dim] = start_id
    else:
        indices[para_dim] = slice(start_id, end_id, None)
    indices = mi[tuple(indices)]

    # Return new_spline
    spline_info = {}
    spline_info["control_points"] = spline_copy.cps[indices]
    spline_info["degrees"] = spline_copy.degrees.tolist()
    if not is_interval:
        spline_info["degrees"].pop(para_dim)
    if spline.has_knot_vectors:
        spline_info["knot_vectors"] = spline_copy.knot_vectors.copy()
        if not is_interval:
            spline_info["knot_vectors"].pop(para_dim)
        else:
            spline_info["knot_vectors"][para_dim] = spline_copy.knot_vectors[
                para_dim
            ][start_id : (end_id + spline_copy.degrees[para_dim] + 1)]

    if spline.is_rational:
        spline_info["weights"] = spline_copy.weights[indices]

    return type(spline)(**spline_info)


def boundaries(spline, boundary_ids=None):
    r"""
    Extracts boundary spline.

    The boundaries deducted from the parametric axis which is normal to the
    boundary (:math:`j`), if the boundary is at parametric axis position
    :math:`x_{j}=x_{j_{min}}` the corresponding boundary is :math:`2j`, else at
    parametric axis position :math:`x_j=x_{j_{min}}` the boundary is
    :math:`2j+1`


    Parameters
    -----------
    spline: Spline / Multipatch
      (`self`-argument if called via extract member of a spline)
    boundary_ids: list
      Only considered for Spline. Default is None and returns all boundaries.

    Returns
    -------
    boundary_spline: type(self)
        boundary spline, which has one less para_dim
    """
    from splinepy import Multipatch as _Multipatch
    from splinepy import splinepy_core as _core

    # Pass to respective c++ implementation
    if isinstance(spline, _Multipatch):
        return spline.boundary_multipatch()
    else:
        bids = [] if boundary_ids is None else list(boundary_ids)
        return [
            type(spline)(spline=c)
            for c in _core.extract_boundaries(spline, bids)
        ]


def arrow_data(spline, adata_name):
    """
    Creates edges that represent arrow_data.
    This function respects entries in show_options.

    Parameters
    ----------
    spline: Spline
    adata_name: str

    Returns
    -------
    adata_edges: gus.Edges
    """
    if adata_name not in spline.spline_data:
        raise KeyError(
            f"given splines does not have arrow data - {adata_name}."
        )

    default_res = 10 if spline.name.startswith("Multi") else 100
    res = spline.show_options.get("resolutions", default_res)
    sampled_spline = _visualize._sample_spline(spline, res)

    adata = _visualize._sample_arrow_data(
        spline, adata_name, sampled_spline, _enforce_len(res, spline.para_dim)
    )

    # if `on` is specified, arrow_data will be a vertices, else it's processed
    # within sampled_spline
    gus_mesh = sampled_spline if adata is None else adata

    arrow_data_value = gus_mesh.vertex_data.as_arrow(adata_name, None, True)

    # if data and origin does not have save dimension, they can't be
    # represented as edges.
    # We can either raise error or return gus_mesh. We will do the latter
    # One can extract origin and value using gus_mesh.vertices and
    # gus_mesh.vertex_data[adata_name]
    if arrow_data_value.shape[1] != spline.dim:
        _log.warning(
            "dimension mismatch between arrow_data and spline."
            "returning sampled mesh as is (not as gus.Edges)."
        )
        return gus_mesh

    as_edges = _from_data(
        gus_mesh,
        arrow_data_value,
        gus_mesh.show_options.get("arrow_data_scale", None),
        data_norm=gus_mesh.vertex_data.as_scalar(adata),
    )

    return as_edges


def bases(spline, ids=None, parametric_view=True):
    """
    Creates spline that represents basis functions of given ids.
    Parametric bounds equals to the given spline.

    Parameters
    ----------
    spline: Spline
    ids: list
      Supports ids of basis functions. Default is None and returns all supports

    Returns
    -------
    basis_splines: list
    """
    # if none, extract all
    if ids is None:
        ids = _np.arange(len(spline.control_points))

    # type check
    if not isinstance(ids, (list, tuple, _np.ndarray)):
        raise TypeError(f"Bases ids ({type(ids)}) should be a list.")

    # first, create conforming parametric view
    if parametric_view:
        conform_para_view = spline.create.parametric_view(
            axes=False, conform=True
        )
        base_basis = conform_para_view.create.embedded(spline.para_dim + 1)
    else:
        base_basis = spline.create.embedded(spline.dim + 1)

    # for template, the last axis should be all zero
    # also, set control point show_option to False
    base_basis.control_points[:, -1] = 0.0
    base_basis.show_options["control_points"] = False

    # loop and e
    basis_splines = []
    for i in ids:
        basis_i = base_basis.copy()
        basis_i.control_points[i, -1] = 1.0
        basis_i.show_options["c"] = int(i)
        basis_splines.append(basis_i)

    return basis_splines


class Extractor:
    """Helper class to allow direct extraction from spline obj (BSpline or
    NURBS). Internal use only.

    Examples
    ---------
    >>> spline_faces = my_spline.extract.faces()
    """

    __slots__ = ("_helpee",)

    def __init__(self, spl):
        from splinepy import Multipatch as _Multipatch
        from splinepy import Spline as _Spline

        if not isinstance(spl, (_Spline, _Multipatch)):
            raise ValueError("Extractor expects a Spline or Multipatch type")
        self._helpee = spl

    @_wraps(edges)
    def edges(self, *args, **kwargs):
        return edges(self._helpee, *args, **kwargs)

    @_wraps(faces)
    def faces(self, *args, **kwargs):
        return faces(self._helpee, *args, **kwargs)

    @_wraps(volumes)
    def volumes(self, *args, **kwargs):
        return volumes(self._helpee, *args, **kwargs)

    @_wraps(control_points)
    def control_points(self):
        return control_points(self._helpee)

    @_wraps(control_edges)
    def control_edges(self):
        return control_edges(self._helpee)

    @_wraps(control_faces)
    def control_faces(self):
        return control_faces(self._helpee)

    @_wraps(control_points)
    def control_volumes(self):
        return control_volumes(self._helpee)

    @_wraps(control_mesh)
    def control_mesh(self):
        return control_mesh(self._helpee)

    def beziers(self):
        """
        Extract all (rational) Bezier patches within a given spline

        Parameters
        ----------
        None

        Returns
        -------
        beziers : list
          List of all individual bezier patches representing the non-zero
          knot-spans
        """
        if not self._helpee.has_knot_vectors:
            return [self._helpee]
        return self._helpee.extract_bezier_patches()

    @_wraps(boundaries)
    def boundaries(self, *args, **kwargs):
        return boundaries(self._helpee, *args, **kwargs)

    def spline(self, splitting_plane=None, interval=None):
        """Extract a spline from a spline.

        Use a (number of) splitting planes to extract a subsection from the
        parametric domain of it.

        Parameters
        ----------
        splitting_plane : int or dictionary {int : (float, float)}
          if int : parametric dimension to be extracted
          if dict : list of splitting planes and ranges to be passed
        interval : float or tuple<float,float>
          interval or value in parametric space to be extracted from the
          spline representation

        Returns
        -------
        spline
        """
        if isinstance(splitting_plane, dict):
            if interval is not None:
                raise ValueError("Arguments incompatible expect dictionary")
            splitting_plane = dict(
                sorted(splitting_plane.items(), key=lambda x: x[0])[::-1]
            )
            spline_copy = self._helpee.copy()
            for key, item in splitting_plane.items():
                spline_copy = spline(spline_copy, key, item)
            return spline_copy
        else:
            return spline(self._helpee, splitting_plane, interval)

    @_wraps(arrow_data)
    def arrow_data(self, *args, **kwargs):
        return arrow_data(self._helpee, *args, **kwargs)

    @_wraps(bases)
    def bases(self, *args, **kwargs):
        return bases(self._helpee, *args, **kwargs)



"""
Helps you find control point ids.
"""

import numpy as _np


class MultiIndex:
    """
    Performs same task as `np.ravel_multi_index`. However, you can use it with
    __getitem__ like queries.
    """

    __slots__ = ("_raveled_indices",)

    def __init__(self, grid_resolutions):
        """
        Initialize using grid's resolution per dimension. For splines, use
        `Spline.control_mesh_resolutions`.

        Parameters
        ----------
        grid_resolutions: array-like
        """
        if not isinstance(grid_resolutions, (tuple, list, _np.ndarray)):
            raise TypeError(
                "grid_resolutions must be array-like type, not"
                f"{type(grid_resolutions)}"
            )

        # create raveled indices
        raveled = _np.arange(_np.prod(grid_resolutions), dtype="int32")

        # to allow general __getitem__ like query, turn them into
        # grid's shape
        self._raveled_indices = raveled.reshape(*grid_resolutions, order="F")

    def __getitem__(self, args):
        """
        Returns raveled indices.

        Parameters
        ----------
        args: int, slice, ellipsis, np.newaxis, array-like
          Any input np.ndarray.__getitem__ would take is valid.

        Returns
        --------
        raveled_indices: np.ndarray
          raveled indices, which are in a raveled array.
        """
        return self._raveled_indices.__getitem__(args).T.ravel()



import gustaf as gus
import numpy as np
import pytest

import splinepy


def test_extract_edges():
    """Test all possible scenarios in the extract edges routine"""
    # 1D Parametric (single)
    bezier_line = splinepy.Bezier(
        degrees=[2], control_points=[[0, 1], [1, 1], [2, 1]]
    )
    edges = bezier_line.extract.edges(resolution=20)
    assert np.allclose(edges.vertices[:, 0], np.linspace(0, 2, 20))

    # 1D Parametric (MP)
    bezier_line_2 = bezier_line.copy()
    bezier_line_2.cps += [2, 0]
    mp_line = splinepy.Multipatch(splines=[bezier_line, bezier_line_2])
    edges = mp_line.extract.edges(resolution=20)
    assert np.allclose(
        edges.vertices[:, 0],
        np.hstack((np.linspace(0, 2, 20), np.linspace(2, 4, 20))),
    )

    # Extract patch-borders (single patch)
    bezier_cube = splinepy.Bezier(
        [1, 1, 1],
        [
            [0, 0, 0],
            [1, 0, 0],
            [0, 1, 0],
            [1, 1, 0],
            [0, 0, 1],
            [1, 0, 1],
            [0, 1, 1],
            [1, 1, 1],
        ],
    ).bspline
    bezier_cube.insert_knots(0, [i * 0.25 for i in range(1, 4)])
    bezier_cube.insert_knots(1, [i * 0.33333 for i in range(1, 3)])
    bezier_cube.insert_knots(2, [i * 0.2 for i in range(1, 5)])
    edges = bezier_cube.extract.edges(resolution=2, all_knots=False)
    expected = np.array(
        [
            [0.0, 0.0, 0.0],
            [1.0, 0.0, 0.0],
            [0.0, 1.0, 0.0],
            [1.0, 1.0, 0.0],
            [0.0, 0.0, 1.0],
            [1.0, 0.0, 1.0],
            [0.0, 1.0, 1.0],
            [1.0, 1.0, 1.0],
            [0.0, 0.0, 0.0],
            [0.0, 1.0, 0.0],
            [1.0, 0.0, 0.0],
            [1.0, 1.0, 0.0],
            [0.0, 0.0, 1.0],
            [0.0, 1.0, 1.0],
            [1.0, 0.0, 1.0],
            [1.0, 1.0, 1.0],
            [0.0, 0.0, 0.0],
            [0.0, 0.0, 1.0],
            [1.0, 0.0, 0.0],
            [1.0, 0.0, 1.0],
            [0.0, 1.0, 0.0],
            [0.0, 1.0, 1.0],
            [1.0, 1.0, 0.0],
            [1.0, 1.0, 1.0],
        ]
    )
    assert np.allclose(edges.vertices, expected)

    # Extract Patch border (MP)
    bezier_cube_2 = bezier_cube.copy()
    bezier_cube_2.cps += [0, 0, 1]
    mp_cubes = splinepy.Multipatch(splines=[bezier_cube, bezier_cube_2])
    edges = mp_cubes.extract.edges(resolution=2, all_knots=False)
    assert np.allclose(
        edges.vertices, np.vstack([expected, expected + [0, 0, 1]])
    )


def test_extract_faces(np_rng):
    """Test extraction of faces"""
    # Two dimensional single patch
    bspline = splinepy.BSpline(
        knot_vectors=[[0, 0, 1, 1], [0, 0, 1, 1]],
        degrees=[1, 1],
        control_points=[[0, 0], [2, 0], [1, 1], [3, 1]],
    )
    # Modify to make more interesting
    bspline.elevate_degrees([0, 0, 1, 1])
    faces = bspline.extract.faces(resolution=6)
    assert np.allclose(faces.vertices, bspline.sample(6))
    # Check a random face center
    face_id = np_rng.integers(0, 5 * 5)
    face_col = face_id // 5
    face_row = face_id % 5
    d_x = 0.2
    query_points = np.array([[0, 0], [d_x, 0], [0, d_x], [d_x, d_x]]) + [
        face_row * d_x,
        face_col * d_x,
    ]
    face_center = np.sum(bspline.evaluate(query_points), axis=0) / 4
    assert np.allclose(face_center, faces.centers()[face_id, :])

    # Two dimensional single patch
    bspline_2 = bspline.copy()
    bspline_2.cps += [1, 1]
    mp_2d = splinepy.Multipatch(splines=[bspline, bspline_2])
    mp_faces = mp_2d.extract.faces(resolution=6, watertight=False)
    assert np.allclose(mp_faces.vertices, mp_2d.sample(6))

    # Check faces
    assert np.allclose(face_center, faces.centers()[face_id, :])

    assert np.allclose(
        face_center + [1, 1],
        mp_faces.centers()[face_id + faces.elements.shape[0], :],
    )

    # Watertighten faces
    mp_faces_wt = mp_2d.extract.faces(resolution=6, watertight=True)

    # Check faces (remain true)
    face_id = np_rng.integers(0, 5 * 5)
    assert np.allclose(
        mp_faces_wt.centers()[face_id, :], faces.centers()[face_id, :]
    )

    assert mp_faces.vertices.shape[0] - mp_faces_wt.vertices.shape[0] == 6

    # p3D
    nurbs = splinepy.NURBS(
        knot_vectors=[[0, 0, 1, 1], [0, 0, 1, 1], [0, 0, 1, 1]],
        degrees=[1, 1, 1],
        control_points=[
            [0, 0, 0],
            [2, 0, 0],
            [1, 1, 0],
            [3, 1, 0],
            [0, 0, 3],
            [2, 0, 3],
            [1, 1, 3],
            [3, 1, 3],
        ],
        weights=np.ones(8),
    )
    faces = nurbs.extract.faces(resolution=8, watertight=False)
    assert np.allclose(faces.vertices.shape[0], 8 * 8 * 6)

    # Check watertighten
    faces = nurbs.extract.faces(resolution=8, watertight=True)
    assert faces.vertices.shape[0] == 8 * 8 * 6 - 6 * 12 - 2 * 8

    # Check multipatch
    nurbs_2 = nurbs.copy()
    nurbs_2.cps += [2, 0, 0]
    mp_3d = splinepy.Multipatch(splines=[nurbs, nurbs_2])
    faces = mp_3d.extract.faces(resolution=8, watertight=False)
    assert np.allclose(faces.vertices.shape[0], 2 * 8 * 8 * 5)

    # Check watertighten
    mp_3d = splinepy.Multipatch(splines=[nurbs, nurbs_2])
    faces = mp_3d.extract.faces(resolution=8, watertight=True)
    assert faces.vertices.shape[0] == 2 * 8 * 8 * 5 - 6 * 20 - 2 * 8 - 3 * 4

    # Check if the boundary centers correspond to the expected faces
    boundaries = nurbs.extract.boundaries() + nurbs_2.extract.boundaries()[1:]
    boundaries.pop(1)  # First id of first spline is duplicate
    expected_faces = gus.Faces.concat(
        [b.extract.faces(resolution=8) for b in boundaries]
    )
    assert np.allclose(
        np.sort(expected_faces.centers(), axis=0),
        np.sort(faces.centers(), axis=0),
    )


def test_extract_volumes():
    """Check the extraction of volumetric meshes from splines"""
    error_spline = splinepy.Bezier(
        degrees=[1, 1], control_points=[[0, 0], [0, 0], [0, 0], [1, 1]]
    )
    with pytest.raises(
        ValueError,
        match=r"Volume extraction from a spline is only valid for para_dim: 3 dim"
        r": 3 splines.",
    ):
        error_spline.extract.volumes(resolution=5)

    # Single Patch
    nurbs = splinepy.NURBS(
        knot_vectors=[[0, 0, 1, 1], [0, 0, 1, 1], [0, 0, 1, 1]],
        degrees=[1, 1, 1],
        control_points=[
            [0, 0, 0],
            [2, 0, 0],
            [1, 1, 0],
            [3, 1, 0],
            [0, 0, 3],
            [2, 0, 3],
            [1, 1, 3],
            [3, 1, 3],
        ],
        weights=np.ones(8),
    )
    volumes_sp = nurbs.extract.volumes(resolution=6)

    # Use linearity of spline to check against evaluations
    queries = splinepy.utils.data.cartesian_product(
        [np.linspace(0.1, 0.9, 5) for _ in range(3)]
    )

    # Check length and compare centers
    assert volumes_sp.elements.shape[0] == 5**3
    assert np.allclose(nurbs.evaluate(queries), volumes_sp.centers())

    # Watertight flag does not have an influence on single patch geometries
    assert np.allclose(
        nurbs.extract.volumes(resolution=6, watertight=True).vertices,
        volumes_sp.vertices,
    )

    # Check multipatch geometries
    nurbs_2 = nurbs.copy()
    nurbs_2.cps += [0, 0, 3]
    mp_nurbs = splinepy.Multipatch(splines=[nurbs, nurbs_2])

    # Retrieve volumes from multipatch
    volumes_mp = mp_nurbs.extract.volumes(resolution=6, watertight=False)

    # Check length and compare centers
    assert volumes_mp.elements.shape[0] == 2 * 5**3
    assert np.allclose(mp_nurbs.evaluate(queries), volumes_mp.centers())

    # Copy mp_nurbs because it currently has issues #198
    mp_nurbs = splinepy.Multipatch(splines=[nurbs, nurbs_2])
    volumes_mp_wt = mp_nurbs.extract.volumes(resolution=6, watertight=True)

    # Watertight flag does not have an influence on single patch geometries
    assert volumes_mp_wt.vertices.shape[0] == 2 * 6**3 - 6**2
    assert np.allclose(volumes_mp.centers(), volumes_mp_wt.centers())


def test_extract_control_points():
    """Extract control points of all splines"""
    bez_el0 = splinepy.Bezier(
        degrees=[1, 1], control_points=[[0, 0], [1, 0], [0, 1], [1, 1]]
    )
    bez_el1 = splinepy.Bezier(
        degrees=[1, 1], control_points=[[1, 0], [2, 0], [1, 1], [2, 1]]
    )

    # Extract ctps from single patches
    ctps0 = bez_el0.extract.control_points()
    ctps1 = bez_el1.extract.control_points()

    # Extract from mp geometry
    mp = splinepy.Multipatch(splines=[bez_el0, bez_el1])
    ctps_mp = mp.extract.control_points()

    assert np.allclose(
        np.vstack([ctps0.vertices, ctps1.vertices]), ctps_mp.vertices
    )


def test_extract_boundaries_2D(are_splines_equal):
    """Create a Spline, extract boundaries and check validity"""
    # uniform
    bez_el0 = splinepy.Bezier(
        degrees=[1, 1], control_points=[[0, 0], [1, 0], [0, 1], [1, 1]]
    )
    boundary_0 = splinepy.Bezier(degrees=[1], control_points=[[0, 0], [0, 1]])
    boundary_3 = splinepy.Bezier(degrees=[1], control_points=[[0, 1], [1, 1]])
    list_of_boundaries = bez_el0.extract.boundaries()
    assert are_splines_equal(list_of_boundaries[0], boundary_0)
    assert are_splines_equal(list_of_boundaries[3], boundary_3)

    # non-uniform
    bez_el0 = splinepy.NURBS(
        knot_vectors=[[0, 0, 0.5, 1, 1], [0, 0, 1, 1]],
        weights=[1.0, 0.5, 1, 1, 0.5, 1],
        degrees=[1, 1],
        control_points=[
            [0, 0],
            [0.5, 0.5],
            [1, 0],
            [0, 1],
            [0.5, 1.5],
            [1, 1],
        ],
    )
    boundary_2 = splinepy.NURBS(
        knot_vectors=[[0, 0, 0.5, 1, 1]],
        weights=[1.0, 0.5, 1],
        degrees=[1],
        control_points=[[0, 0], [0.5, 0.5], [1, 0]],
    )
    list_of_boundaries = bez_el0.extract.boundaries()
    assert are_splines_equal(list_of_boundaries[2], boundary_2)


def test_extract_boundaries_3D(are_splines_equal):
    """Create a Spline, extract boundaries and check validity"""
    # uniform
    bez_el0 = splinepy.Bezier(
        degrees=[1, 1, 1],
        control_points=[
            [0, 0, 0],
            [1, 0, 0],
            [0, 1, 0],
            [1, 1, 0],
            [0, 0, 1],
            [1, 0, 1],
            [0, 1, 1],
            [1, 1, 1],
        ],
    )
    boundary_0 = splinepy.Bezier(
        degrees=[1, 1],
        control_points=[
            [0, 0, 0],
            [0, 1, 0],
            [0, 0, 1],
            [0, 1, 1],
        ],
    )
    list_of_boundaries = bez_el0.extract.boundaries()[0]
    assert are_splines_equal(list_of_boundaries, boundary_0)


def test_extract_arrow_data(np_rng):
    """
    see if arrow_data is created as we wished
    """
    spline = splinepy.helpme.create.box(10, 20, 30)
    # try to avoid zero division below
    spline.cps -= 0.0001

    # set self data and resolutions
    spline.spline_data["me"] = spline
    res = [11, 13, 17]
    spline.show_options["resolutions"] = res

    # 1. check number of edges
    me_edges = spline.extract.arrow_data("me")
    assert isinstance(me_edges, gus.Edges)
    # 3d should give you arrows on the faces
    total = 0
    for i in range(spline.para_dim):
        r = res.copy()
        r.pop(i)
        total += np.prod(r) * 2

    assert len(me_edges.edges) == total

    # 2. check if they point to the direction they need to point
    def check_values(a, b):
        # normalize - this may do some zero division
        a /= np.linalg.norm(a, axis=1).reshape(-1, 1)
        b /= np.linalg.norm(b, axis=1).reshape(-1, 1)
        a[np.isnan(a)] = 0
        b[np.isnan(b)] = 0

        assert np.allclose(a, b)

    direction = np.diff(me_edges.vertices[me_edges.edges], axis=1).reshape(
        -1, me_edges.vertices.shape[1]
    )
    sample = spline.extract.faces(res, watertight=False).vertices

    check_values(direction, sample)

    # 3. do the same with locations
    locations = np_rng.random((10, 3))
    spline.spline_data["me_at"] = splinepy.SplineDataAdaptor(
        spline, locations=locations
    )
    me_at_edges = spline.extract.arrow_data("me_at")

    # 3.1 len test
    assert isinstance(me_at_edges, gus.Edges)
    assert len(me_at_edges.edges) == len(locations)

    # 3.2 arrow validity
    direction = np.diff(
        me_at_edges.vertices[me_at_edges.edges], axis=1
    ).reshape(-1, me_at_edges.vertices.shape[1])
    sample = spline.evaluate(locations)

    check_values(direction, sample)

    # 4. at last, scaling
    spline.show_options["arrow_data_scale"] = 200
    me_at_edges = spline.extract.arrow_data("me_at")

    # 3.1 len test
    assert isinstance(me_at_edges, gus.Edges)
    assert len(me_at_edges.edges) == len(locations)

    # 3.2 arrow validity
    direction = np.diff(
        me_at_edges.vertices[me_at_edges.edges], axis=1
    ).reshape(-1, me_at_edges.vertices.shape[1])
    sample = spline.evaluate(locations)
    assert np.allclose(
        direction, sample * spline.show_options["arrow_data_scale"]
    )


def test_extract_spline():
    """Spline from spline extraction"""
    spline = splinepy.Bezier(
        degrees=[4],
        control_points=[
            [0, 0],
            [1, 0],
            [1, 1],
            [2, 1],
            [2, 2],
        ],
    )

    queries = np.linspace(0, 1, 4).reshape(-1, 1)
    for interval in [[0.0, 0.3], [0.2, 0.5], [0.6, 1.0]]:
        new_spline = spline.extract.spline(0, interval)
        linear_projection = queries * (interval[1] - interval[0]) + interval[0]
        assert np.allclose(
            spline.evaluate(linear_projection),
            new_spline.evaluate(queries),
        )

    spline = splinepy.Bezier(
        degrees=[2, 2],
        control_points=[
            [0, 0],
            [2, 1],
            [4, 0],
            [0, 2],
            [2, 3],
            [4, 2],
            [0, 4],
            [2, 5],
            [4, 4],
        ],
    )
    queries = splinepy.utils.data.cartesian_product(
        [np.linspace(0, 1, 4) for _ in range(2)]
    )

    for axis in [0, 1]:
        for interval in [[0.0, 0.3], [0.2, 0.5], [0.6, 1.0], 0.0, 0.4, 1.0]:
            new_spline = spline.extract.spline(axis, interval)
            linear_projection = queries.copy()
            if isinstance(interval, list):
                linear_projection[:, axis] = (
                    linear_projection[:, axis] * (interval[1] - interval[0])
                    + interval[0]
                )
                new_spline_queries = queries.copy()
            else:
                linear_projection[:, axis] = interval
                new_spline_queries = queries[:, (axis + 1) % 2].reshape(-1, 1)

            assert np.allclose(
                spline.evaluate(linear_projection),
                new_spline.evaluate(new_spline_queries),
            )


@pytest.mark.parametrize(
    "spline",
    (
        "rational_bezier_2p2d",
        "bezier_2p2d",
        "bspline_2p2d",
        "nurbs_2p2d",
        "rational_bezier_3p3d",
        "bezier_3p3d",
        "bspline_3p3d",
        "nurbs_3p3d",
    ),
)
def test_extract_bases(spline, request):
    # skip 3d splines for minimal builds
    if splinepy.splinepy_core.is_minimal() and "3d" in spline:
        return None

    spline = request.getfixturevalue(spline)

    # prepare query
    res = 10
    query_res = [res] * spline.para_dim
    q = splinepy.utils.data.uniform_query(spline.parametric_bounds, query_res)

    # prepare reference values
    basis, support = spline.basis_and_support(q)
    mat = np.zeros((len(q), len(spline.cps)))
    np.put_along_axis(mat, support, basis, axis=1)

    # get bases
    bases = spline.extract.bases()

    for i, b in enumerate(bases):
        assert np.allclose(b.evaluate(q)[:, -1], mat[:, i])



import numpy as np

import splinepy

if __name__ == "__main__":
    """
    x1
    *---*---*
    | 2 | 3 |
    *---*---*
    | 0 | 1 |
    *---*---* -> x0
    """
    box0 = splinepy.helpme.create.box(1, 2)
    x0_max, x1_max = box0.cps.max(axis=0)

    box1 = box0.copy()
    box1.cps[:, 0] += x0_max

    box2 = box0.copy()
    box2.cps[:, 1] += x1_max

    box3 = box0.copy()
    box3.cps += [x0_max, x1_max]

    # plot scalar data
    m = splinepy.Multipatch([box0, box1, box2, box3])
    m.spline_data["t"] = m
    m.show_options["data"] = "t"
    m.show_options["lighting"] = "off"
    m.show()

    # plot arrow data
    m.show_options["arrow_data"] = "t"
    m.show()

    # plot arrow data with callback
    def plot_eval(data, on):
        """
        evaluates at given location
        """
        return data.evaluate(on)

    def plot_jacs(data, on):
        """
        evaluates at given location
        """
        return np.linalg.det(np.vstack([d.jacobian(on) for d in data.patches]))

    # use adaptor to plot values at specific places
    # this works because multipatch's evaluate() evaluates at
    # each spline. you need to make sure that your queries are within
    # each bounds.
    m.spline_data["t2"] = splinepy.SplineDataAdaptor(m, function=plot_eval)
    m.show_options["arrow_data"] = "t2"

    # form query and set evaluation points for arrow_data
    circle = splinepy.helpme.create.circle(radius=0.3)
    circle.cps -= circle.cps.min(axis=0)
    queries = circle.sample(20)[:-1]
    m.show_options["arrow_data_on"] = queries
    m.show()

    # Generate MS
    generator = splinepy.microstructure.Microstructure()
    generator.deformation_function = splinepy.Bezier(
        degrees=[2, 1],
        control_points=[[0, 0], [2, 1], [4, 0], [0, 2], [2, 4], [4, 2]],
    )
    generator.microtile = splinepy.microstructure.tiles.Cross2D()
    generator.tiling = [3, 2]

    m = generator.create(center_expansion=1.2)
    m.spline_data["detJ"] = splinepy.SplineDataAdaptor(m, function=plot_jacs)
    m.show_options["data"] = "detJ"
    m.show_options["lighting"] = "off"
    m.show()



"""
Abstract Spline
"""

import copy as _copy
import os as _os

import numpy as _np

from splinepy import helpme as _helpme
from splinepy import io as _io
from splinepy import settings as _settings
from splinepy import splinepy_core as _core
from splinepy import utils as _utils
from splinepy._base import SplinepyBase as _SplinepyBase
from splinepy.helpme import visualize as _visualize
from splinepy.helpme.check import Checker as _Checker
from splinepy.helpme.create import Creator as _Creator
from splinepy.helpme.extract import Extractor as _Extractor
from splinepy.helpme.integrate import Integrator as _Integrator
from splinepy.utils.data import SplineData as _SplineData
from splinepy.utils.data import cartesian_product as _cartesian_product


class RequiredProperties(_SplinepyBase):
    """
    Helper class to hold required properties of each spline.
    """

    # name mangling to make direct access very annoying
    __required_spline_properties = {
        "Bezier": ("degrees", "control_points"),
        "RationalBezier": ("degrees", "control_points", "weights"),
        "BSpline": ("degrees", "knot_vectors", "control_points"),
        "NURBS": ("degrees", "knot_vectors", "control_points", "weights"),
    }

    # union of all req props. nice for check support
    __union_all = {
        rp for rps in __required_spline_properties.values() for rp in rps
    }

    # intersection of all req props. nice for checking minimal support
    __intersection_all = __union_all.intersection(
        *[set(rps) for rps in __required_spline_properties.values()]
    )

    @classmethod
    def __call__(cls, spline):
        """
        Given splinepy.Spline, returns required properties.
        Wraps `__getitem__`.

        Parameters
        -----------
        spline: Spline or type

        Returns
        --------
        required_spline_properties: list
        """
        if isinstance(spline, str):
            return cls.__getitem__(spline)

        # extract pure `type`
        s_type = spline if isinstance(spline, type) else type(spline)

        # __getitem__ checks if it is supported
        return cls.__getitem__(s_type.__qualname__)

    @classmethod
    def __getitem__(cls, spline_type):
        """
        Given a name of spline type in str, returns required properties.
        Checks if given spline type is supported.

        Parameters
        -----------
        spline_type: str

        Returns
        --------
        required_spline_properties: list
        """
        # do we support this spline type?
        if spline_type not in cls.__required_spline_properties:
            raise ValueError(
                f"Sorry, we don't have support for ({spline_type})-types."
                "Supported spline types are:"
                f"{list(cls.__required_spline_properties.keys())}"
            )

        return cls.__required_spline_properties[spline_type]

    @classmethod
    def of(cls, spline):
        """
        Wrapper of __call__ and __getitem__ as classmethod.

        Parameters
        -----------
        spline: Spline or str

        Returns
        --------
        required_spline_properties: list

        Examples
        ---------
        >>> myspline = NURBS(...)
        >>> requirements = RequiredSplineProperties.of(myspline)
        >>> requirements
        ['degrees', 'knot_vectors', 'control_points', 'weights']
        """
        if isinstance(spline, str):
            return cls.__getitem__(spline)
        else:
            return cls.__call__(spline)

    @classmethod
    def union(cls, *splines):
        """
        Checks union of required properties of given splines.
        If nothing is given, returns union of all supported splines.

        Parameters
        ----------
        *splines: Spline or str

        Returns
        -------
        rp_union: list
        """
        if len(splines) == 0:
            return cls.__union_all

        rp_union = set()
        for s in splines:
            rp_union.update(cls.of(s))

        return rp_union

    @classmethod
    def intersection(cls, *splines):
        """
        Checks intersection of required properties of given splines.
        If nothing is given, returns intersection of all supported splines.

        Parameters
        ----------
        *splines: Spline or str

        Returns
        -------
        rp_intersection: list
        """
        if len(splines) == 0:
            return cls.__intersection_all

        rp_intersection = set(cls.of(splines[-1]))
        for s in splines[:-1]:
            rp_intersection.intersection_update(cls.of(s))

        return rp_intersection


def _default_if_none(arg, default):
    """
    Returns default if arg is None

    Parameters
    ----------
    arg: object
    default: object

    Returns
    -------
    default_if_none: object
    """
    return default if arg is None else arg


def _prepare_coordinates(spl):
    """
    Prepares physical space array. Internally called when saved control points
    are  not physicalspacearrays.

    Parameters
    ----------
    spl: Spline

    Returns
    -------
    None
    """
    if not _core.has_core(spl):
        return None

    ws = True  # dummy init
    cps = spl._data.get("control_points", None)
    if spl.is_rational:
        ws = spl._data.get("weights", None)

    # make sure we have all the properties we need
    if cps is None or ws is None:
        current_properties = spl.current_core_properties()
        cps = current_properties["control_points"]
        if spl.is_rational:
            ws = current_properties["weights"]

    # get coordinate pointers
    coordinate_pointers = spl._coordinate_pointers()

    cps = cps.view(_utils.data.PhysicalSpaceArray)
    cps._source_ptr = coordinate_pointers[0]
    spl._data["control_points"] = cps

    if spl.is_rational:
        ws = ws.view(_utils.data.PhysicalSpaceArray)
        ws._source_ptr = coordinate_pointers[1]
        spl._data["weights"] = ws.reshape(-1, 1)


def _get_helper(spl, attr_name, helper_class):
    """
    Returns a helper object from given spline. If it doesn't exist, it will
    create one first.

    Parameters
    ----------
    spl: Spline
    attr_name: str
    helper_class: type

    Returns
    -------
    helper: object
    """
    attr = getattr(spl, attr_name, None)
    if attr is not None:  # hasattr
        return attr

    # if attr doesn't exist, create one and save to its member defined in
    # slots
    helper_obj = helper_class(spl)
    setattr(spl, attr_name, helper_obj)

    return helper_obj


def _safe_new_core(spl, exclude="?"):
    """
    Calls all getters of spline's required_properties.
    This guarantees that spline's saved data (spl._data) contains local
    copy/reference of core spline's properties, as getters properly processes
    and saves spline properties.

    This is useful to call within the setters before calling spl._new_core.
    For example, spline manipulation calls like spl.elevate_degrees() deletes
    all the saved data as most of the saved data is invalid afterwards.
    If you then call a setter, it will set a new value and naturally try to
    create a new spline using cached properties in spl._data, which is
    incomplete at this point.

    Parameters
    ----------
    spl: Spline
    exclude: str
      Default is "?" as no member name can start with a question mark.

    Returns
    -------
    None
    """
    # make sure spl._data has all the properties we need
    if _core.has_core(spl):
        for rp in spl.required_properties:
            if not rp.startswith(exclude):
                _ = getattr(spl, rp, None)

    # try to sync with current status
    spl._new_core(
        keep_properties=True,
        raise_=False,
        **spl._data,
    )


def _safe_array_copy(spl, array, key, coupled_key):
    """
    In splinepy we make safe copies. For control_points and weights,
    to support non-contiguous / homogeneous control points in backend,
    we have a helper class called `PhysicalSpaceArray`. This syncs
    numpy array and backend's control points through `ControlPointPointers`.

    This function checks a non-copy criteria: self assignment.
    This tends to happen during inplace operations.
    For example, spline.cps += 1 becomes
    cp_ref = spline.cps; cp_ref +=1; spline.cps = cp_ref.
    In this case, we don't want to create a new core spline.

    Parameters
    ----------
    spl: Spline
    array: np.ndarray
    key: str
    coupled_key: str

    Returns
    -------
    copied: bool
      Whether copy was made
    """
    # is current value same as given array?
    if spl._data.get(key, None) is array:
        return False

    coupled_array = spl._data.get(coupled_key, None)
    if coupled_array is not None:
        # length check - they need to match
        if len(coupled_array) != len(array):
            raise ValueError(
                f"len({key}) ({len(array)}) "
                f"should match len({coupled_key}) ({len(coupled_array)})."
            )

        # for rational splines, control points and weights are coupled
        # -> in backends, we save weighted control points and they must
        #    belong to a same spline.
        # for example, if we are setting new control points,
        # we need to create a copy of weights so that these new control points
        # are coupled with a fresh weight array.
        #
        # Else, the following assert will raise
        # >>> orig_sample = spl.sample(n)
        # >>> spl.cps = new_cps
        # >>> spl.ws[[1,2,3]] *= .5
        # >>> assert not np.allclose(orig_sample, spl.sample(n))
        if isinstance(coupled_array, _utils.data.PhysicalSpaceArray):
            spl._data[coupled_key] = coupled_array.copy()

    # safety copy
    spl._data[key] = _np.array(array, dtype="float64", copy=True, order="C")
    return True


class Spline(_SplinepyBase, _core.PySpline):
    r"""
    Spline base class. Extends :class:`.PySpline` with documentation.

    Generally, all types of splines can be seen as mappings from a
    :math:`N_{param}`-dimensional parametric domain
    :math:`\Omega_{param} \subset \mathbb{R}^{N_{param}}` to a
    :math:`N_{phys}`-dimensional physical domain
    :math:`\Omega_{phys} \subset \mathbb{R}^{N_{phys}}`, i.e.,

    .. math::
            M : \Omega_{param} \to \Omega_{phys}

    :code:`splinepy` generally supports different combinations of embeddings,
    that is different combinations of dimensionalities :math:`N_{param}` and
    :math:`N_{phys}`, up to dimension 10.
    The supported spline types differ in the way how the spline function
    :math:`M` is defined. For an overview of the mathematical theories for the
    different types of supported splines, we refer to the documentation of the
    classes :class:`.BSplineBase` and :class:`.BezierBase` for a more in-depth
    discussion of the theory on B-Spline/Bezier families, as well as to their
    children :class:`.BSpline`, :class:`.NURBS`, :class:`.Bezier`, and
    :class:`.RationalBezier` for usage examples.

    Parameters
    -----------
    spline: Spline
        Initialize using another spline. Will SHARE core spline.
    degrees: (para_dim,) array-like
        Keyword only parameter.
    knot_vectors: (para_dim, n) list of array-like
        Keyword only parameter.
    control_points: (m, dim) array-like
        Keyword only parameter.
    weights: (m,) array-like
        Keyword only parameter.

    Returns
    -------
    None
    """

    __slots__ = (
        "_show_options",
        "_spline_data",
    )

    __show_option__ = _visualize.SplineShowOption

    def __init__(self, spline=None, **kwargs):
        # return if this is an empty init
        if spline is None and len(kwargs) == 0:
            super().__init__()
            return None

        # current core spline based init if given spline has no local changes
        if spline is not None and isinstance(spline, _core.PySpline):
            # spline based init - takes SplinepyBase, even the nullptr
            super().__init__(spline)
            # this should keep all the cps and kv references if there's any
            # make shallow copy
            self._data = spline._data.copy()
            return None

        else:
            super().__init__()  # alloc

            # use setters to check minimal requirements and enforce
            # contiguous array
            # in depth checks are done in core
            req_properties = self.required_properties
            if type(self).__qualname__.startswith("Spline"):
                # for Spline class, which can take anything,
                # we want to set it in this specific order that
                # it won't keep creating core
                set_order = (
                    "weights",
                    "control_points",
                    "knot_vectors",
                    "degrees",
                )
            else:
                set_order = req_properties

            for to_set in set_order:
                if to_set in kwargs:
                    setattr(self, to_set, kwargs[to_set])

    @property
    def required_properties(self):
        """
        Returns required property of current spline.

        Parameters
        -----------
        None

        Returns
        --------
        requied_properties: dict
        """
        try:
            return RequiredProperties.of(self)
        except ValueError:
            # Spline
            self._logd(
                f"`required_properties` of {type(self)} is undefined."
                "Returning maximal set of properties."
            )
            return RequiredProperties.union()

    @property
    def para_dim(self):
        """
        Returns parametric dimension.

        Parameters
        -----------
        None

        Returns
        --------
        para_dim: int
        """
        return super().para_dim

    @property
    def dim(self):
        """
        Returns physical dimension.

        Parameters
        -----------
        None

        Returns
        --------
        dim: int
        """
        return super().dim

    @property
    def whatami(self):
        """
        Answers a deep philosophical question of "what am i?"

        Parameters
        -----------
        None

        Returns
        --------
        whatami: str
        """
        return super().whatami

    @property
    def name(self):
        """
        Name of the spline provided by cpp side as a str.
        Should be same as type(spline).__qualname__ if you are using
        a specified spline, i.e., not `Spline`.

        Parameters
        -----------
        None

        Returns
        -------
        name: str
        """
        return super().name

    @property
    def has_knot_vectors(self):
        """
        Returns True iff spline has knot vectors. Bezier splines don't.

        Parameters
        ----------
        None

        Returns
        -------
        has_knot_vectors: bool
        """
        return super().has_knot_vectors

    @property
    def is_rational(self):
        """
        Returns True iff spline is rational. NURBS is rational, for example.

        Parameters
        ----------
        None

        Returns
        -------
        is_rational: bool
        """
        return super().is_rational

    @property
    def extract(self):
        """Returns spline extractor. Can directly perform extractions available
        at `splinepy/helpme/extract.py`.

        Examples
        ---------
        >>> spline_faces = spline.extract.faces()

        Parameters
        -----------
        None

        Returns
        --------
        extractor: Extractor
        """
        return _Extractor(self)

    @property
    def check(self):
        """Returns spline checker. Can directly perform type and value checks
        at `splinepy/helpme/check.py`.

        Examples
        ---------
        >>> spline.check.valid_queries(queries)

        Parameters
        -----------
        None

        Returns
        --------
        checker: Checker
        """
        return _Checker(self)

    @property
    def integrate(self):
        """Returns spline integrator. Can directly perform integrations
        available at `splinepy/helpme/integrate.py`.

        Examples
        ---------

        .. code-block :: python

          spline_faces = spline.integrate.volume()

        Parameters
        -----------
        None

        Returns
        --------
        integrator: Integrator
        """
        return _Integrator(self)

    @property
    def create(self):
        """Returns spline creator Can be used to create new splines using
        geometric relations.

        Examples
        --------
        >>> prism = spline.create.extrude(axis=[1, 4, 1])

        Parameters
        ----------
        None

        Returns
        -------
        creator: spline.Creator
        """
        return _Creator(self)

    @property
    def show_options(self):
        """
        Show option manager for splines.

        Parameters
        ----------
        None

        Returns
        -------
        show_options: SplineShowOption
        """
        return _get_helper(self, "_show_options", self.__show_option__)

    @property
    def spline_data(self):
        """
        Spline data helper for splines.

        Parameters
        ----------
        None

        Returns
        -------
        spline_data: SplineData
        """
        return _get_helper(self, "_spline_data", _SplineData)

    def clear(self):
        """
        Clears core spline and saved data

        Parameters
        ----------
        None

        Returns
        -------
        None
        """
        # annulment of core
        _core.annul_core(self)
        self._data = {}

    def _new_core(self, *, keep_properties=False, raise_=True, **kwargs):
        """
        Creates a new core spline based on given kwargs.
        Clears any saved data afterwards.

        Parameters
        -----------
        keep_properties: bool
          Default is False. If True, keeps current properties. Otherwise,
          all internal temporary data will be deleted, including current
          properties.
        raise_: bool
          Default is True. Raises if kwargs don't contain required
          properties.
        **kwargs: kwargs
          Keyword only argument. Takes properties.

        Returns
        -------
        core_created: bool
        """
        # let's remove None valued items and make values contiguous array
        kwargs = _utils.data.enforce_contiguous_values(kwargs)

        # type check for knot vectors
        kvs = kwargs.get("knot_vectors", None)
        if kvs is not None:
            kwargs["knot_vectors"] = [
                kv.numpy() if isinstance(kv, _core.KnotVector) else kv
                for kv in kvs
            ]

        # Spline, you do whatever
        if type(self).__qualname__ == "Spline":
            # maybe minimal set check?
            super()._new_core(**kwargs)

        # specified ones needs specific sets of kwargs
        # in case of an incomplete set of kwargs, nothing will happen
        elif set(kwargs.keys()).issuperset(set(self.required_properties)):
            rp = self.required_properties
            rp_dict = {}
            for k, v in kwargs.items():
                if k in rp:
                    rp_dict[k] = v

            super()._new_core(**rp_dict)

        elif raise_:
            raise RuntimeError(
                f"{kwargs.keys()} is not valid set of keywords to create "
                f"a new core for {type(self).__qualname__}."
                f"Valid ones should include: {self.required_properties}."
            )

        else:
            self._logd(
                "Couldn't initialize a new_core with given keywords."
                f"Keywords without None values are {kwargs.keys()}."
            )

        # clear saved data
        if not keep_properties:
            # BSpline supports viewing-control-points for contiguous arrays
            # such as np.ndarray, so we need to keep it alive.
            if self.name.startswith("BSpline"):
                saved_cps = self._data.get("control_points", None)
                self._data = {}
                if saved_cps is not None:
                    self._logw(
                        "_new_core(keep_properties=False) -",
                        "BSplines need to keep control_points.",
                        "Properties excluding control_points will be cleared.",
                    )
                    self._data["control_points"] = saved_cps
            else:
                self._data = {}

        return _core.has_core(self)

    @property
    def degrees(self):
        """
        Returns Degrees.

        Parameters
        -----------
        None

        Returns
        --------
        degrees: (para_dim,) np.ndarray
        """
        ds = self._data.get("degrees", None)

        if _core.has_core(self):
            if ds is None:
                ds = self._current_core_degrees()
                ds.flags.writeable = False
                self._data["degrees"] = ds
            else:
                ds.flags.writeable = False

        return ds

    @degrees.setter
    def degrees(self, degrees):
        """
        Sets Degrees.

        Parameters
        -----------
        degrees: list-like

        Returns
        --------
        None
        """
        # clear core and return
        if degrees is None:
            _core.annul_core(self)
            self._data = {}
            return None

        # len should match knot_vector's len
        if self.knot_vectors is not None and len(self.knot_vectors) != len(
            degrees
        ):
            raise ValueError(
                f"len(degrees) ({len(degrees)}) should match "
                f"len(self.knot_vectors) ({len(self.knot_vectors)})."
            )

        # set - copies
        self._data["degrees"] = _np.asarray(
            degrees, dtype="int32", order="C"
        ).copy()

        # make sure _new_core call works
        _safe_new_core(self, exclude="degrees")

    @property
    def knot_vectors(self):
        """
        Returns knot vectors.

        Parameters
        -----------
        None

        Returns
        --------
        knot_vectors: list
        """
        kvs = self._data.get("knot_vectors", None)
        if _core.has_core(self):
            if not self.has_knot_vectors:
                return None

            if not isinstance(kvs, _core.ParameterSpace):
                kvs = self._parameter_space()
                self._data["knot_vectors"] = kvs

        return kvs

    @knot_vectors.setter
    def knot_vectors(self, knot_vectors):
        """
        Set knot vectors.

        Parameters
        -----------
        knot_vectors: list

        Returns
        --------
        None
        """
        # clear core and return
        if knot_vectors is None:
            _core.annul_core(self)
            self._data = {}
            return None

        # len should match degrees's len
        if self.degrees is not None and len(knot_vectors) != len(self.degrees):
            raise ValueError(
                f"len(knot_vectors) ({len(knot_vectors)}) should "
                f"match len(self.degrees) ({len(self.degrees)})."
            )

        # set - copies
        new_kvs = []
        for kv in knot_vectors:
            if isinstance(kv, _core.KnotVector):
                new_kvs.append(kv.numpy())
            else:
                new_kvs.append(_utils.data.enforce_contiguous(kv, "float64"))

        self._data["knot_vectors"] = new_kvs

        _safe_new_core(self, exclude="knot_vectors")

    @property
    def unique_knots(self):
        """
        Returns unique knots.
        Does not store results.

        Parameters
        -----------
        None

        Returns
        --------
        unique_knots: list
        """
        if not self.has_knot_vectors:
            self._logd(
                "Returning parametric_bounds as "
                "Bezier spline's unique knots."
            )
            return list(self.parametric_bounds.T)

        else:
            self._logd("Retrieving unique knots")
            return self.knot_vectors.unique_knots()

    @property
    def knot_multiplicities(self):
        """
        Returns knot multiplicities.
        Does not store results.

        Parameters
        -----------
        None

        Returns
        --------
        knot_multiplicities: list<np.array>
        """
        if "Bezier" in self.name:
            self._logd(
                "Returning multiplicities of knots if Bezier was BSpline"
            )
            return [_np.array([d + 1, d + 1]) for d in self.degrees]

        else:
            self._logd("Retrieving knot multiplicities")
            return [k.multiplicities() for k in self.knot_vectors]

    @property
    def parametric_bounds(self):
        """
        Returns bounds of parametric_space.

        Parameters
        -----------
        None

        Returns
        --------
        parametric_bounds: (2, para_dim) np.ndarray
        """
        return super().parametric_bounds

    @property
    def control_points(self):
        """
        Returns control points.

        Parameters
        -----------
        None

        Returns
        --------
        control_points_: (n, dim) np.ndarray
        """
        cps = self._data.get("control_points", None)

        if _core.has_core(self) and not isinstance(
            cps, _utils.data.PhysicalSpaceArray
        ):
            _prepare_coordinates(self)

        return self._data.get("control_points", None)

    @control_points.setter
    def control_points(self, control_points):
        """
        Set control points.

        Parameters
        -----------
        control_points: (n, dim) array-like

        Returns
        --------
        None
        """
        # clear core and return
        if control_points is None:
            _core.annul_core(self)
            self._data = {}
            return None

        # set - copies if it is not the same value
        if not _safe_array_copy(
            self, control_points, "control_points", "weights"
        ):
            return None

        _safe_new_core(self, exclude="control_points")

    @property
    def control_point_bounds(self):
        """
        Returns bounds of control points.

        Parameters
        -----------
        None

        Returns
        --------
        control_point_bounds: (2, dim) np.ndarray
        """
        cps = self.control_points

        return _np.vstack((cps.min(axis=0), cps.max(axis=0)))

    @property
    def control_mesh_resolutions(self):
        """
        Returns control mesh resolutions.

        Parameters
        -----------
        None

        Returns
        --------
        control_mesh_resolutions: (para_dim) np.ndarray
        """
        return super().control_mesh_resolutions

    def mapper(self, reference):
        """Retrieve a mapper that can be used to get physical derivatives such
        as a gradient or hessian in physical space

        Parameters
        ----------
        reference : spline
          Spline that represents the geometry of the field

        Returns
        -------
        mapper : Mapper
          Mapper to calculate physical gradients and hessians
        """
        return _helpme.mapper.Mapper(self, reference=reference)

    def greville_abscissae(self, duplicate_tolerance=None):
        """
        Returns greville abscissae.

        Parameters
        -----------
        duplicate_tolerance : float
          If negative, duplicates are possible for C^(-1) splines (B-Spline
          family). Otherwise they are filtered out using adjacent abscissae.
          Value represents tolerance between two greville abscissae, to be
          considered as equal

        Returns
        --------
        greville_abscissae: (para_dim) np.ndarray
        """
        return _cartesian_product(
            super()._greville_abscissae_list(
                _default_if_none(duplicate_tolerance, -1.0),
            )
        )

    @property
    def multi_index(self):
        """
        Easy control point / weights access using (unraveled) multi index.
        Useful for getting indices if you want to "slice" a control mesh,
        or find ids that relate to a certain boundary.

        Parameters
        ----------
        None

        Returns
        -------
        multi_index_helper: MultiIndex
        """
        mi = self._data.get("multi_index", None)
        if mi is not None:
            return mi

        # not stored, create one
        # this is okay to be copied together in copy()
        self._data["multi_index"] = _helpme.multi_index.MultiIndex(
            self.control_mesh_resolutions
        )

        return self._data["multi_index"]

    @property
    def weights(self):
        """
        Returns weights.

        Parameters
        -----------
        None

        Returns
        --------
        self._weights: (n, 1) array-like
        """
        ws = self._data.get("weights", None)

        if _core.has_core(self) and not isinstance(
            ws, _utils.data.PhysicalSpaceArray
        ):
            _prepare_coordinates(self)

        return self._data.get("weights", None)

    @weights.setter
    def weights(self, weights):
        """
        Set weights.

        Parameters
        -----------
        weights: (n,) array-like

        Returns
        --------
        None
        """
        # clear core and return
        if weights is None:
            _core.annul_core(self)
            self._data = {}
            return None

        # set - copies if it is not the same value
        # if it is the same, return
        if not _safe_array_copy(self, weights, "weights", "control_points"):
            return None

        _safe_new_core(self, exclude="weights")

    def evaluate(self, queries, nthreads=None):
        """
        Evaluates spline.

        Parameters
        -----------
        queries: (n, para_dim) array-like
        n_threads: int

        Returns
        --------
        results: (n, dim) np.ndarray
        """
        self._logd("Evaluating spline")

        queries = _utils.data.enforce_contiguous(
            queries, dtype="float64", asarray=_settings.CHECK_BOUNDS
        )

        if _settings.CHECK_BOUNDS:
            self.check.valid_queries(queries)

        return super().evaluate(
            queries,
            nthreads=_default_if_none(nthreads, _settings.NTHREADS),
        )

    def sample(self, resolutions, nthreads=None):
        """
        Uniformly sample along each parametric dimensions from spline.

        Parameters
        -----------
        resolutions: (n,) array-like
        nthreads: int

        Returns
        --------
        results: (math.product(resolutions), dim) np.ndarray
        """
        from gustaf.utils import arr as _arr

        resolutions = _arr.enforce_len(resolutions, self.para_dim)

        self._logd(f"Sampling {_np.prod(resolutions)} points from spline.")

        return super().sample(
            resolutions,
            nthreads=_default_if_none(nthreads, _settings.NTHREADS),
        )

    def derivative(self, queries, orders, nthreads=None):
        """
        Evaluates derivatives of spline.

        Parameters
        -----------
        queries: (n, para_dim) array-like
        orders: (para_dim,) or (m, para_dim) array-like
        nthreads: int

        Returns
        --------
        results: (n, m, dim) np.ndarray
          Iff m == 1, it will have (n, dim) shape.
        """
        self._logd("Evaluating derivatives of the spline")

        queries = _utils.data.enforce_contiguous(
            queries, dtype="float64", asarray=_settings.CHECK_BOUNDS
        )

        if _settings.CHECK_BOUNDS:
            self.check.valid_queries(queries)

        orders = _utils.data.enforce_contiguous(orders, dtype="int32")

        return super().derivative(
            queries=queries,
            orders=orders,
            nthreads=_default_if_none(nthreads, _settings.NTHREADS),
        )

    def jacobian(self, queries, nthreads=None):
        """
        Evaluates jacobians on spline.

        Parameters
        -----------
        queries: (n, para_dim) array-like
        n_threads: int

        Returns
        --------
        results: (n, para_dim, dim) np.ndarray
        """
        self._logd("Determining spline jacobians")

        queries = _utils.data.enforce_contiguous(
            queries, dtype="float64", asarray=_settings.CHECK_BOUNDS
        )

        if _settings.CHECK_BOUNDS:
            self.check.valid_queries(queries)

        return super().jacobian(
            queries,
            nthreads=_default_if_none(nthreads, _settings.NTHREADS),
        )

    def support(self, queries, nthreads=None):
        """
        Returns basis support ids of given queries.

        Parameters
        -----------
        queries: (n, para_dim) array-like
        n_threads: int

        Returns
        --------
        support: (n, prod(degrees + 1)) np.ndarray
        """
        self._logd("Evaluating support ids")
        queries = _utils.data.enforce_contiguous(
            queries, dtype="float64", asarray=_settings.CHECK_BOUNDS
        )

        if _settings.CHECK_BOUNDS:
            self.check.valid_queries(queries)

        return super().support(
            queries=queries,
            nthreads=_default_if_none(nthreads, _settings.NTHREADS),
        )

    def basis(self, queries, nthreads=None):
        """
        Returns basis function values on the supports of given queries.

        Parameters
        -----------
        queries: (n, para_dim) array-like
        n_threads: int

        Returns
        --------
        basis: (n, prod(degrees + 1)) np.ndarray
        """
        self._logd("Evaluating basis functions")
        queries = _utils.data.enforce_contiguous(
            queries, dtype="float64", asarray=_settings.CHECK_BOUNDS
        )

        if _settings.CHECK_BOUNDS:
            self.check.valid_queries(queries)

        return super().basis(
            queries=queries,
            nthreads=_default_if_none(nthreads, _settings.NTHREADS),
        )

    def basis_and_support(self, queries, nthreads=None):
        """
        Returns basis function values and their support ids of given queries.
        Same as calling `basis` and `support` at the same time.

        Parameters
        -----------
        queries: (n, para_dim) array-like
        n_threads: int

        Returns
        --------
        basis: (n, prod(degrees + 1)) np.ndarray
        support: (n, prod(degrees + 1)) np.ndarray
        """
        self._logd("Evaluating basis functions and support")
        queries = _utils.data.enforce_contiguous(
            queries, dtype="float64", asarray=_settings.CHECK_BOUNDS
        )

        if _settings.CHECK_BOUNDS:
            self.check.valid_queries(queries)

        return super().basis_and_support(
            queries=queries,
            nthreads=_default_if_none(nthreads, _settings.NTHREADS),
        )

    def basis_derivative(self, queries, orders, nthreads=None):
        """
        Returns derivative of basis functions evaluated on the supports
        of given queries.

        Parameters
        ----------
        queries: (n, para_dim) array-like
        orders: (para_dim,) or (m, para_dim) array-like
        nthreads: int

        Returns
        --------
        basis_derivatives: (n, m, prod(degrees + 1)) np.ndarray
          Iff m == 1, it will have (n, prod(degrees + 1)) shape.
        """
        self._logd("Evaluating basis function derivatives")
        queries = _utils.data.enforce_contiguous(
            queries, dtype="float64", asarray=_settings.CHECK_BOUNDS
        )

        if _settings.CHECK_BOUNDS:
            self.check.valid_queries(queries)

        orders = _utils.data.enforce_contiguous(orders, dtype="int32")

        return super().basis_derivative(
            queries=queries,
            orders=orders,
            nthreads=_default_if_none(nthreads, _settings.NTHREADS),
        )

    def basis_derivative_and_support(self, queries, orders, nthreads=None):
        """
        Returns derivative of basis functions and their support ids of given
        queries. Same as calling `basis_derivative` and `support` at the same
        time.

        Parameters
        ----------
        queries: (n, para_dim) array-like
        orders: (para_dim,) or (m, para_dim) array-like
        nthreads: int

        Returns
        --------
        basis_derivatives: (n, m, prod(degrees + 1)) np.ndarray
          Iff m == 1, it will have (n, prod(degrees + 1)) shape.
        supports: (n, prod(degrees + 1)) np.ndarray
        """
        self._logd("Evaluating basis function derivatives")
        queries = _utils.data.enforce_contiguous(
            queries, dtype="float64", asarray=_settings.CHECK_BOUNDS
        )

        if _settings.CHECK_BOUNDS:
            self.check.valid_queries(queries)

        orders = _utils.data.enforce_contiguous(orders, dtype="int32")

        return super().basis_derivative_and_support(
            queries=queries,
            orders=orders,
            nthreads=_default_if_none(nthreads, _settings.NTHREADS),
        )

    def proximities(
        self,
        queries,
        initial_guess_sample_resolutions=None,
        tolerance=None,
        max_iterations=-1,
        tight_search_bounds=False,
        nthreads=None,
        return_verbose=False,
    ):
        """
        Given physical coordinate, finds a parametric coordinate that maps to
        the nearest physical coordinate. Also known as "point inversion".

        Makes initial guess by searching for the nearest points with kd-tree.
        With `initial_guess_sample_resolutions` value,
        physical points will be sampled to plant a kd-tree.

        The planted kd-tree is saved in cpp object internally
        and will re-plant if `initial_guess_sample_resolutions` are positive.
        Default behavior is to re-plant the tree each query.
        If you know that spline doesn't have any inplace changes, and would
        like to avoid it, set negative values.

        If there's no tree, this will raise runtime error.

        Parameters
        -----------
        queries: (n, dim) array-like
        initial_guess_sample_resolutions: (para_dim) array-like
          Default is 2 * control_mesh_resolutions. For complex shapes, we
          recommend setting large values.
        tolerance: float
          Convergence criteria. Currently for both distance and residual
        max_iterations: int
          Default is (para_dim * 20)
        tight_search_bounds: bool
          Default is False.
          Set search bound to aabb of direct neighbors from the sample grid.
        nthreads: int
        return_verbose : bool
          If False, returns only parametric coords

        Returns
        --------
        para_coord: (n, para_dim) np.ndarray
          Parametric coordinates
        phys_coord: (n, dim) np.ndarray
          (only if return_verbose) respective physical coordinates
        phys_diff: (n, dim) np.ndarray
          (only if return_verbose) respective cartesian difference to query
        distance: (n, 1) np.ndarray
          (only if return_verbose) respective 2-norm of difference
        convergence_norm: (n, 1) np.ndarrray
          (only if return_verbose) Newton residual
        first_derivatives: (n, para_dim, dim) np.ndarray
          (only if return_verbose) Convergence information
        second_derivatives: (n, para_dim, para_dim, dim) np.ndarray
          (only if return_verbose) Convergence information
        """
        self._logd("Searching for nearest parametric coord")

        queries = _utils.data.enforce_contiguous(queries, dtype="float64")

        # set small tolerance.
        if tolerance is None and _settings.TOLERANCE > 1.0e-14:
            tolerance = 1e-14

        verbose_info = super().proximities(
            queries=queries,
            initial_guess_sample_resolutions=_default_if_none(
                initial_guess_sample_resolutions,
                self.control_mesh_resolutions * 2,
            ),
            tolerance=tolerance,
            max_iterations=max_iterations,
            tight_search_bounds=tight_search_bounds,
            nthreads=_default_if_none(nthreads, _settings.NTHREADS),
        )

        if return_verbose:
            return verbose_info
        else:
            if _np.any(
                verbose_info[4]
                > _default_if_none(tolerance, _settings.TOLERANCE)
            ):
                self._logw(
                    "Proximity search did not converge within the tolerance "
                    "for some queries. Try to rerun proximity search with "
                    "return_verbose=True to get more information."
                )
            return verbose_info[0]

    def elevate_degrees(self, parametric_dimensions):
        """
        Elevate degree.

        Parameters
        -----------
        parametric_dimensions: list

        Returns
        --------
        None
        """
        if _settings.CHECK_BOUNDS:
            self.check.clamped_knot_vectors(warning=True)

        super().elevate_degrees(para_dims=parametric_dimensions)
        self._logd(
            f"Elevated {parametric_dimensions}.-dim. " "degree of the spline."
        )
        self._data = {}

    def reduce_degrees(self, parametric_dimensions, tolerance=None):
        """
        Tries to reduce degree.

        Parameters
        -----------
        parametric_dimensions: list
        tolerance: float

        Returns
        --------
        reduced: list
        """
        if _settings.CHECK_BOUNDS:
            self.check.clamped_knot_vectors(warning=True)

        reduced = super().reduce_degrees(
            para_dims=parametric_dimensions,
            tolerance=_default_if_none(tolerance, _settings.TOLERANCE),
        )

        def meaningful(r):
            """True/False to reduced/failed"""
            return "reduced" if r else "failed"

        self._logd(
            f"Tried to reduce degrees for {parametric_dimensions}.-dims. "
            "Results: ",
            f"{[meaningful(r) for r in reduced]}.",
        )

        if any(reduced):
            self._data = {}

        return reduced

    def export(self, fname):
        """
        Export spline. Please be aware of the limits of `.iges`.
        vtk exports are excluded here because it is not a true spline export.

        Parameters
        -----------
        fname: str

        Returns
        --------
        None
        """
        # prepare export destination
        fname = str(fname)
        dirname = _os.path.dirname(fname)
        if not _os.path.isdir(dirname) and dirname != "":
            _os.makedirs(dirname)

        # ext tells you what you
        ext = _os.path.splitext(fname)[1]

        if ext == ".iges":
            _io.iges.export(fname, [self])

        elif ext == ".cats":
            _io.cats.export(fname.replace(".cats", ".xml"), [self])

        elif ext == ".xml":
            _io.gismo.export(fname, [self])

        elif ext == ".itd":
            _io.irit.export(fname, [self])

        elif ext == ".npz":
            _io.npz.export(fname, self)

        elif ext == ".json":
            _io.json.export(fname, self)

        elif ext == ".mesh":
            # mfem nurbs mesh.
            _io.mfem.export(fname, self, precision=12)

        else:
            raise ValueError(
                "We can only export "
                "< .iges | .xml | .itd | .npz | .mesh | .json> extensions"
            )

        self._logi(f"Exported current spline as {fname}.")

    def todict(self, tolist=False):
        """
        Return current spline as dict. Copies all the dict values.
        Does not check current status.

        Parameters
        -----------
        tolist: bool
          Default is False. Converts `np.ndarray` into lists.

        Returns
        --------
        dict_spline: dict
        """
        self._logd("Preparing dict_spline")
        if _core.has_core(self):
            core_properties = self.current_core_properties()
            if tolist:
                new_dict = {}
                for k, v in core_properties.items():
                    if k.startswith("knot_vectors"):
                        new_dict[k] = [vv.tolist() for vv in v]
                    else:
                        new_dict[k] = v.tolist()
                return new_dict
            else:
                return core_properties

        dict_spline = {}
        # loop and copy entries.
        for p in self.required_properties:
            # no prop? no prob. default is None
            tmp_prop = getattr(self, p, None)
            should_copy = True
            # attr are either list or np.ndarray
            # prepare list if needed.
            if tolist and (tmp_prop is not None):
                if isinstance(tmp_prop, _np.ndarray):
                    tmp_prop = tmp_prop.tolist()  # copies
                    should_copy = False
                if p.startswith("knot_vectors"):
                    tmp_prop = [list(tp) for tp in tmp_prop]
                    should_copy = False

            if should_copy:
                tmp_prop = _copy.deepcopy(tmp_prop)

            # update
            dict_spline[p] = tmp_prop

        return dict_spline

    def copy(self, saved_data=True):
        """
        Returns deepcopy of stored data and newly initialized self.
        Shallow copies are made for `spline_data`.

        Parameters
        -----------
        saved_data: bool
          Default is True. calls deepcopy on saved data

        Returns
        --------
        new_spline: type(self)
        """
        new = type(self)(**self.current_core_properties())

        # new's _data is populated with current_core_properties(), which they
        # own. So we will add to that
        if saved_data:
            required_properties = self.required_properties
            for k, v in self._data.items():
                if k in required_properties:
                    continue
                new._data[k] = _copy.deepcopy(v)

            # copy show_options and shallow copy spline_data
            self.show_options.copy_valid_options(new.show_options)
            new.spline_data._saved = self.spline_data._saved.copy()

        return new

    def show(self, **kwargs):
        """Equivalent to

        .. code-block:: python

            splinepy.helpme.visualize.show(spline, **kwargs)

        Parameters
        ----------
        kwargs: kwargs
          see splinepy.helpme.visualize.show

        Returns
        -------
        showable_or_plotter: dict or vedo.Plotter
        """
        return _visualize.show(self, **kwargs)

    def showable(self, **kwargs):
        """Equivalent to

        .. code-block:: python

            splinepy.helpme.visualize.show(
                spline, return_showable=True, **kwargs
            )

        Parameters
        ----------
        kwargs: kwargs
          see splinepy.helpme.visualize.show

        Returns
        -------
        spline_showable: dict
        """
        return _visualize.show(self, return_showable=True, **kwargs)

    # short cuts / alias
    ds = degrees
    kvs = knot_vectors
    cps = control_points
    ws = weights



"""splinepy/utils/data.py.

Helps helpee to manage data. Some useful data structures.
"""

from itertools import accumulate as _accumulate
from itertools import chain as _chain
from itertools import repeat as _repeat

import numpy as _np
from gustaf.helpers.data import DataHolder as _DataHolder
from gustaf.utils.arr import enforce_len as _enforce_len

from splinepy import splinepy_core as _splinepy_core
from splinepy._base import SplinepyBase as _SplinepyBase

try:
    import scipy as _scipy

    has_scipy = True
except ImportError:
    has_scipy = False

__all__ = [
    "PhysicalSpaceArray",
    "_DataHolder",
    "enforce_contiguous_values",
    "cartesian_product",
    "SplineDataAdaptor",
    "SplineData",
]


class PhysicalSpaceArray(_np.ndarray):
    """numpy array object that keeps mirroring inplace changes to the source.
    Meant to help control_points.
    """

    __slots__ = (
        "_source_ptr",
        "_row_indices",
        "_full_row_indices",
        "_super_arr",
    )

    def __array_finalize__(self, obj):
        """Sets default flags for any arrays that maybe generated based on
        physical space array. For more information,
        see https://numpy.org/doc/stable/user/basics.subclassing.html"""
        self._source_ptr = None
        self._super_arr = None

        # for arrays created based on this subclass
        if isinstance(obj, type(self)):
            # this is copy. nothing to worry here
            if self.base is None:
                return None

            # first child array
            if self.base is obj:
                # make sure this is not a recursively born child
                # for example, `arr[[1,2]][:,2]`
                if obj._source_ptr is not None:
                    self._super_arr = obj
                return None

            # multi generation child array
            if obj._super_arr is not None and self.base is obj.base:
                self._super_arr = obj._super_arr
                return None

            return None

    def _sync_source_ptr(self):
        # this is super arr. super arr only has _source_ptr because we should've
        # set it manually
        if self._source_ptr is not None:  # and self._super_arr is None
            self._source_ptr.sync(self)
            return None

        # this is a sub arr viewing super arr
        # this will perform a full sync
        if self._super_arr is not None and self._source_ptr is None:
            self._super_arr._source_ptr.sync(self._super_arr)
            return None

    def copy(self, *args, **kwargs):
        """copy creates regular numpy array"""
        return _np.array(self, *args, copy=True, **kwargs)

    def view(self, *args, **kwargs):
        """Set writeable flags to False for the view."""
        v = super(self.__class__, self).view(*args, **kwargs)
        v.flags.writeable = False
        return v

    def __iadd__(self, *args, **kwargs):
        sr = super(self.__class__, self).__iadd__(*args, **kwargs)
        self._sync_source_ptr()
        return sr

    def __isub__(self, *args, **kwargs):
        sr = super(self.__class__, self).__isub__(*args, **kwargs)
        self._sync_source_ptr()
        return sr

    def __imul__(self, *args, **kwargs):
        sr = super(self.__class__, self).__imul__(*args, **kwargs)
        self._sync_source_ptr()
        return sr

    def __idiv__(self, *args, **kwargs):
        sr = super(self.__class__, self).__idiv__(*args, **kwargs)
        self._sync_source_ptr()
        return sr

    def __itruediv__(self, *args, **kwargs):
        sr = super(self.__class__, self).__itruediv__(*args, **kwargs)
        self._sync_source_ptr()
        return sr

    def __imatmul__(self, *args, **kwargs):
        sr = super(self.__class__, self).__imatmul__(*args, **kwargs)
        self._sync_source_ptr()
        return sr

    def __ipow__(self, *args, **kwargs):
        sr = super(self.__class__, self).__ipow__(*args, **kwargs)
        self._sync_source_ptr()
        return sr

    def __imod__(self, *args, **kwargs):
        sr = super(self.__class__, self).__imod__(*args, **kwargs)
        self._sync_source_ptr()
        return sr

    def __ifloordiv__(self, *args, **kwargs):
        sr = super(self.__class__, self).__ifloordiv__(*args, **kwargs)
        self._sync_source_ptr()
        return sr

    def __ilshift__(self, *args, **kwargs):
        sr = super(self.__class__, self).__ilshift__(*args, **kwargs)
        self._sync_source_ptr()
        return sr

    def __irshift__(self, *args, **kwargs):
        sr = super(self.__class__, self).__irshift__(*args, **kwargs)
        self._sync_source_ptr()
        return sr

    def __iand__(self, *args, **kwargs):
        sr = super(self.__class__, self).__iand__(*args, **kwargs)
        self._sync_source_ptr()
        return sr

    def __ixor__(self, *args, **kwargs):
        sr = super(self.__class__, self).__ixor__(*args, **kwargs)
        self._sync_source_ptr()
        return sr

    def __ior__(self, *args, **kwargs):
        sr = super(self.__class__, self).__ior__(*args, **kwargs)
        self._sync_source_ptr()
        return sr

    def __array_wrap__(self, *args, **kwargs):
        sr = super(self.__class__, self).__array_wrap__(*args, **kwargs)
        self._sync_source_ptr()
        return sr

    def __setitem__(self, key, value):
        # set first. invalid setting will cause error
        sr = super(self.__class__, self).__setitem__(key, value)

        # for child array, we will just sync and exit.
        if self._source_ptr is None:
            self._sync_source_ptr()
            return sr

        # if set item above worked.
        # then, mirror / copy rowwise
        # if this was called due to inplace operations of subset of entries
        # this would be redundant
        #
        # here, we will try to parse very simple, yet, quite often used
        # indexings
        contig = _np.ascontiguousarray  # frequently used
        if isinstance(key, int) or (
            isinstance(key, _np.ndarray) and key.ndim == 0
        ):
            # single get item is always contiguous
            self._source_ptr.set_row(key, self[key])

        elif isinstance(key, _np.ndarray) and key.ndim == 1:
            if key.dtype.kind.startswith("b"):
                key = contig(self.row_indices()[key])

            self._source_ptr.sync_rows(contig(key), self)

        elif isinstance(key, list):
            if isinstance(key[0], bool):
                key = contig(self.row_indices()[key])
            self._source_ptr.sync_rows(key, self)

        # tuple is a bit tricky
        elif isinstance(key, tuple) and (
            isinstance(key[0], int)
            or (isinstance(key[0], _np.ndarray) and key[0].ndim == 0)
        ):
            self._source_ptr.set_row(key[0], self[key[0]])

        elif isinstance(key, tuple) and (
            isinstance(key[0], (list, _np.ndarray))
        ):
            if isinstance(key[0][0], (_np.bool_, bool)):
                key = (contig(self.row_indices()[key[0]]),)

            self._source_ptr.sync_rows(key[0], self)

        elif isinstance(key, slice) or (
            isinstance(key, tuple) and isinstance(key[0], slice)
        ):
            key = key if isinstance(key, slice) else key[0]

            start = 0 if key.start is None else key.start
            # flip negatives
            start = start + len(self) if start < 0 else start
            stop = len(self) if key.stop is None else key.stop
            # flip negatives
            stop = stop + len(self) if stop < 0 else stop
            step = 1 if key.step is None else key.step

            ids = _np.arange(
                start, stop, step, dtype="int32"
            )  # should be contig
            self._source_ptr.sync_rows(ids, self)

        elif isinstance(key, type(Ellipsis)) or (
            isinstance(key, tuple) and isinstance(key[0], type(Ellipsis))
        ):
            self._source_ptr.sync(self)

        else:
            # even more complex? then
            # TODO - just sync if this seems to take too long
            ids = _np.unique(self.full_row_indices()[key])
            self._source_ptr.sync_rows(contig(ids), self)

        return sr

    def row_indices(self):
        """
        Returns row_indices. Same results as np.arange(len(arr))
        """
        if hasattr(self, "_row_indices"):
            return self._row_indices

        len_, d = self._source_ptr.len(), self._source_ptr.dim()
        self._row_indices = _np.arange(len_, dtype="int32")
        self._full_row_indices = self._row_indices.repeat(d).reshape(len_, d)
        return self._row_indices

    def full_row_indices(self):
        """
        Returns full_row_indices. Same results as np.indices(arr.shape)[0]
        """
        if hasattr(self, "_full_row_indices"):
            return self._full_row_indices

        len_, d = self._source_ptr.len(), self._source_ptr.dim()
        self._row_indices = _np.arange(len_, dtype="int32")
        self._full_row_indices = self._row_indices.repeat(d).reshape(len_, d)
        return self._full_row_indices


def enforce_contiguous(array, dtype=None, asarray=False):
    """
    If input is an instance / subclass of np.ndarray, this will check
    if they are configuous. If so, returns same object, else turns makes it
    contiguous and returns.

    Parameters
    ----------
    array: array-like
      data to be transformed
    dtype : type
      base data type (default float)
    asarray : bool
      also transform list or tuple to np.ndarray

    Returns
    -------
    contiguous_array: array-like
    """
    if isinstance(array, _np.ndarray):
        if array.flags["C_CONTIGUOUS"] and (
            dtype is None or _np.dtype(dtype) is array.dtype
        ):
            return array
        return _np.ascontiguousarray(array, dtype=dtype)

    if asarray and isinstance(array, (list, tuple, _splinepy_core.KnotVector)):
        return _np.ascontiguousarray(array, dtype=dtype)

    return array


def enforce_contiguous_values(dict_):
    """
    Returns a new dict where values are contiguous. If the value is
    an array, enforces contiguous elements for one recursive level.
    This also removes None values.

    Parameters
    ----------
    dict_: dict

    Returns
    -------
    dict_with_contiguous: dict
    """
    dict_with_contiguous = {}
    for key, value in dict_.items():
        # most likely will be asking about np.ndarray
        if isinstance(value, _np.ndarray):
            contig_value = enforce_contiguous(value)

        # or, list. only check if the first element is np.ndarray
        elif isinstance(value, list) and isinstance(value[0], _np.ndarray):
            contig_value = [enforce_contiguous(v) for v in value]

        # abandon Nones
        elif value is None:
            continue

        else:
            contig_value = value

        # set item
        dict_with_contiguous[key] = contig_value

    return dict_with_contiguous


def cartesian_product(arrays, reverse=True):
    """
    Fast Cartesian product.
    Adapted from stackoverflow entry:
    https://stackoverflow.com/questions/11144513/
    cartesian-product-of-x-and-y-array-points-into-single-array-of-2d-points/
    49445693#49445693
    Answer from `Paul Panzer`.

    In splinepy, we use the ordering, where lowest dimension changes fastest.
    Thus, a reverse option.

    Parameters
    ----------
    arrays: tuple or list
      tuple or list of array-like
    reverse: bool
      Default is True. Reverses order of fastest iterating dimension.
      If False, highest indexed dimension iterates fastest.

    Returns
    -------
    cartesian: (n, len(arrays)) np.ndarray
    """
    n_arr = len(arrays)
    dtype = _np.result_type(*arrays)

    # for reverse, use view of reverse ordered arrays
    if reverse:
        arrays = arrays[::-1]

    shape = *map(len, arrays), n_arr
    cartesian = _np.empty(shape, dtype=dtype)

    # reverse order the view of output array, so that reverse
    # is again in original order, but just reversed fastest iterating dim
    _cartesian = cartesian[..., ::-1] if reverse else cartesian

    dim_reducing_views = (
        *_accumulate(
            _chain((_cartesian,), _repeat(0, n_arr - 1)),
            _np.ndarray.__getitem__,
        ),
    )
    idx = slice(None), *_repeat(None, n_arr - 1)
    for i in range(n_arr - 1, 0, -1):
        dim_reducing_views[i][..., i] = arrays[i][idx[: n_arr - i]]
        dim_reducing_views[i - 1][1:] = dim_reducing_views[i]
    _cartesian[..., 0] = arrays[0][idx]

    return cartesian.reshape(-1, n_arr)


def make_matrix(values, supports, n_cols, as_array=False):
    r"""
    Create a matrix from values and supports.

    Used for basis functions their derivatives and everything mapped.
    Uses scipy if available. If `as_array` is true, dense matrix (numpy) is
    enforced

    This matrix can be used for approximations and IGA-applications. With given
    queries :math:`\pmb{\xi}`, control points :math:`\mathbf{C}`, associated to
    spline :math:`\mathcal{S}`, this function can be used to return matrix
    :math:`\mathbf{A}`, such that

    .. math::
        \mathcal{S}(\pmb{\xi}) = A(\pmb{\xi}) \cdot \mathbf{C}

    Parameters
    ----------
    values : np.ndarray
      Values to be inserted into the matrix
    supports : np.ndarray
      Column values corresponding to the given values
    n_cols : int
      Number of columns of the matrix (matches n_ctps for basis-functions)
    as_array : bool
      Return as numpy / dense type

    Returns
    -------
    matrix : np.ndarray / scipy.sparse.csr_array
      Matrix
    """
    # Return in Matrix shape (scipy if available)
    if has_scipy and not as_array:
        return _scipy.sparse.csr_array(
            (
                values.ravel(),
                (
                    _np.arange(values.shape[0]).repeat(values.shape[1]),
                    supports.ravel(),
                ),
            ),
            shape=(values.shape[0], n_cols),
        )
    else:
        matrix = _np.zeros((values.shape[0], n_cols))
        _np.put_along_axis(matrix, supports, values, axis=1)
        return matrix


def uniform_query(bounds, resolutions):
    """
    Creates uniform query within the given bounds and resolutions.
    Same as `gus.create.vertices.raster(bounds, resolution).vertices`.

    Parameters
    ----------
    bounds: 2D array-like
      [[lower_1, lower_2, ...], [upper_1, upper_2, ...]]
    resolutions: 1D array-like
      [resolution_1, resolution_2, ...]
    """
    # unpack bounds
    lower_b, upper_b = bounds

    if not (len(lower_b) == len(upper_b) == len(resolutions)):
        raise ValueError(
            "lower and upper bounds should have same len as resolutions"
        )

    # create per-dimension queries
    queries_per_dim = []
    for lb, ub, r in zip(lower_b, upper_b, resolutions):
        queries_per_dim.append(_np.linspace(lb, ub, r))

    return cartesian_product(queries_per_dim, reverse=True)


class SplineDataAdaptor(_SplinepyBase):
    """
    Prepares data to be presentable on spline. To support both
    scalar-data and vector-data, which are representable with colors and
    arrows respectively, this class will prepare data accordingly.

    Parameters
    ----------
    data: any
      Any data that you want to plot on to the spline. If this is not a
      spline, it requires a `function` that takes this data to create
      appropriate values.
    locations: 2D array-like
      Optional. If specified, used to evaluate data at specific locations.
      Applicable for arrow data
    function: callable
      Optional. A callable used to evaluate values using data at query points.
      function(data, queries) -> values
    parametric_bounds: 2D array-like
      Optional. Parametric bounds of supporting spline. Required for stand-alone
      use. Otherwise, set by helping spline. Used to compute query points for
      resolution based sampling.
    """

    __slots__ = (
        "data",
        "function",
        "locations",
        "is_spline",
        "has_function",
        "has_locations",
        "has_evaluate",
        "arrow_data_only",
        "parametric_bounds",
        "_user_created",
    )

    def __init__(
        self, data, locations=None, function=None, parametric_bounds=None
    ):
        """ """
        from splinepy import Multipatch as _Multipatch
        from splinepy import Spline as _Spline

        # default
        self._user_created = True
        self.data = data
        self.is_spline = False
        self.has_function = False
        self.has_locations = False
        self.has_evaluate = False
        self.arrow_data_only = False
        self.parametric_bounds = parametric_bounds

        # is spline we know?
        if isinstance(data, (_Spline, _Multipatch)):
            self.is_spline = True

        # data has evaluate?
        if hasattr(data, "evaluate"):
            self.has_evaluate = callable(data.evaluate)

        # has function?
        if function is not None:
            self.has_function = True
            if not callable(function):
                raise ValueError("Given function isn't callable")
            self.function = function

        # locations? - keep this compatible with functions. maybe
        # we want to have some state dependent value at certain locations
        if locations is not None:
            # set what holds true
            self.has_locations = True
            self.arrow_data_only = True
            self.locations = _np.asanyarray(locations)

            # if this is not a spline we know, it doesn't have a function,
            # it should:
            # -> `data.evaluate` is callable, or
            # -> len(data) == len(locations)
            if not self.is_spline and not self.has_function:
                len_matches = False
                if hasattr(data, "__len__"):
                    len_matches = len(locations) == len(data)
                if not (self.has_evaluate or len_matches):
                    raise ValueError(
                        "Data cannot be represented at specified locations."
                        "Requires one of the following requirements: "
                        "1) is a spline derived from splinepy's spline; "
                        "2) data has `data.evaluate()`; "
                        "3) length of the data and location should match."
                    )
            # location is specified, meaning we don't need sample()
            return None

        # can call sample or has a function?
        if not self.has_function and not self.is_spline:
            raise ValueError(
                "None spline data should at least have an accompanying "
                "function."
            )

    def as_vertex_data(self, resolutions=None, on=None):
        """
        Parameters
        ----------
        resolutions: list or tuple
        at: (n, d) array-like

        Returns
        -------
        vertex_data: (m, r) array-like
        """
        if resolutions is not None and on is not None:
            raise ValueError(
                "Please only specify either `resolutions` or `on`"
            )

        if self.has_locations and (resolutions is not None or on is not None):
            raise ValueError(
                "Location dependent data can't be evaluated with `resolutions`"
                " or `at`."
            )

        # if resolutions is specified, this is not a location query
        if resolutions is not None:
            if self.has_function:
                # parametric bounds needs to be set by now.
                queries = uniform_query(
                    self.parametric_bounds,
                    _enforce_len(resolutions, len(self.parametric_bounds[0])),
                )
                # pass to evaluate.
                return self.function(
                    self.data,
                    on=queries,
                )
            elif self.is_spline and self.data.para_dim > 2:
                # TODO: replace this with generalized query helpers.
                return self.data.extract.faces(
                    resolutions, watertight=False
                ).vertices
            else:
                return self.data.sample(resolutions)

        # runtime location query
        if on is not None:
            if self.has_function:
                return self.function(self.data, on=on)
            elif self.has_evaluate:
                return self.data.evaluate(on)
            else:
                raise ValueError(
                    "Given data can't support data extraction on specified "
                    f"locations ({on})."
                )

        # location specified - either evaluate function at the
        if self.has_locations:
            if self.has_function:
                # function may want locations
                try:
                    return self.function(self.data, self.locations)
                except TypeError:  # maybe too many args
                    return self.function(self.data)
            elif self.has_evaluate:
                return self.data.evaluate(self.locations)
            else:
                return self.data

        # should be returned by now
        raise RuntimeError("Something went wrong while preparing spline data.")


class SplineData(_DataHolder):
    """
    Data manager for splines.
    """

    __slots__ = ()

    def __init__(self, helpee):
        """ """
        if "Spline" not in str(type(helpee).__mro__):
            raise AttributeError("Helpee is not a Spline")

        super().__init__(helpee)

    def __setitem__(self, key, value):
        """
        Selectively accept spline data.

        Parameters
        ----------
        key: str
        value: object

        Returns
        -------
        None
        """
        if isinstance(value, SplineDataAdaptor):
            self._saved[key] = value
        else:
            adapted = SplineDataAdaptor(value)  # will test usability
            adapted._user_created = False  # mark for __getitem__
            self._saved[key] = adapted

        # set parametric_bounds
        self._saved[key].parametric_bounds = self._helpee.parametric_bounds

    def __getitem__(self, key):
        """
        Returns value from __setitem__

        Parameters
        ----------
        key: str

        Returns
        -------
        value: object
        """
        saved = super().__getitem__(key)
        if saved._user_created:
            return saved
        else:
            return saved.data

    def as_scalar(self, key, resolutions, default=None):
        """
        Return scalar value at given resolutions

        Parameters
        ----------
        key: str
        resolutions: list or tuple
        default: object
          Default is None and will return is key doesn't exist

        Returns
        -------
        value: np.ndarray
        """
        if key not in self._saved:
            return default

        saved = super().__getitem__(key)
        # will raise
        return saved.as_vertex_data(resolutions=resolutions)

    def as_arrow(self, key, resolutions=None, on=None, default=None):
        """
        Returns as-arrow-representable data on certain places, with given
        resolution, or on predefined places.

        Parameters
        ----------
        key: str
        resolutions: list or tuple
        on: array-like
        """
        if key not in self._saved:
            return default

        saved = super().__getitem__(key)
        # will raise
        return saved.as_vertex_data(resolutions=resolutions, on=on)


class MultipatchData(SplineData):
    """
    Data manager for multipatch
    """

    __slots__ = ()

    def __init__(self, helpee):
        """
        Strictly, we don't need to inherit from SplineData and only implement
        interfaces, since _saved equivalent is managed in multipatch.
        However, it's an index based holder and we want to support the same
        syntax.
        Here, we will use _saved as key_to_id lookup
        """
        if "Multipatch" not in str(type(helpee).__mro__):
            raise AttributeError("Helpee is not a multipatch")

        _DataHolder.__init__(self, helpee)

    def __setitem__(self, key, value):
        """
        Parameters
        ----------
        key: str
        value: object

        Returns
        -------
        None
        """
        if isinstance(value, SplineDataAdaptor):
            self._saved[key] = value
            # set parametric_bounds of the first patch
            value.parametric_bounds = self._helpee.patches[0].parametric_bounds
        elif "PyMultipatch" in str(type(value).__mro__):
            # get id of this field that's about to be added
            self._saved[key] = len(self._helpee.fields)
            # add fields expects list of list
            # TODO @jzwar add type check to except both?
            self._helpee.add_fields([value.patches], value.dim)
        elif isinstance(value, int):
            # also accept integers as a reference to the field
            self._saved[key] = value
        else:
            raise TypeError(
                "MultipatchData supports SplineDataAdapter or Multipatch."
                f"given type: {type(value)}"
            )

    def __getitem__(self, key):
        """
        Returns value from __setitem__

        Parameters
        ----------
        key: str

        Returns
        -------
        value: object
        """

        # integer input refers to saved field's number
        if isinstance(key, int):
            return self._helpee.fields[key]
        elif isinstance(key, str):
            saved = _DataHolder.__getitem__(self, key)  # will raise KeyError
            if isinstance(saved, int):
                return self._helpee.fields[saved]
            elif isinstance(saved, SplineDataAdaptor):
                return saved
        raise RuntimeError(
            "Invalid saved data / key type. Please help us by writing an "
            "issue at github.com/tataratat/splinepy. Thank you!"
        )

    def as_scalar(self, key, resolutions, default=None):
        """
        Return scalar value at given resolutions

        Parameters
        ----------
        key: str
        resolutions: list or tuple
        default: object
          Default is None and will return is key doesn't exist

        Returns
        -------
        value: np.ndarray
        """
        if key not in self._saved:
            return default

        saved = self.__getitem__(key)
        if isinstance(saved, SplineDataAdaptor):
            # explicitly checks that saved IS data adaptor
            pass
        else:
            # or will create one on the fly
            saved = SplineDataAdaptor(saved)

        return saved.as_vertex_data(resolutions=resolutions)

    def as_arrow(self, key, resolutions=None, on=None, default=None):
        """
        Returns as-arrow-representable data on certain places, with given
        resolution, or on predefined places.

        Parameters
        ----------
        key: str
        resolutions: list or tuple
        on: array-like
        """
        if key not in self._saved:
            return default

        saved = self.__getitem__(key)
        if isinstance(saved, SplineDataAdaptor):
            # explicitly checks that saved IS data adaptor
            pass
        else:
            # or will create one on the fly
            saved = SplineDataAdaptor(saved)

        return saved.as_vertex_data(resolutions=resolutions, on=on)



import numpy as np
import pytest

import splinepy

# frequently used fixtures
all_2p2d_splines = (
    "rational_bezier_2p2d",
    "bezier_2p2d",
    "bspline_2p2d",
    "nurbs_2p2d",
)

all_splines = (
    "rational_bezier_2p2d",
    "bezier_2p2d",
    "bspline_2p2d",
    "nurbs_2p2d",
    "rational_bezier_3p3d",
    "bezier_3p3d",
    "bspline_3p3d",
    "nurbs_3p3d",
)


def test_create_embedded(bspline_2p2d, nurbs_2p2d):
    """
    Test embedding for (rational) splines
    """

    embedded_bspline = bspline_2p2d.create.embedded(3)
    assert np.allclose(
        bspline_2p2d.cps[:, :2],
        embedded_bspline.cps[:, :2],
    )
    embedded_nurbs = nurbs_2p2d.create.embedded(3)
    assert np.allclose(
        nurbs_2p2d.cps[:, :2],
        embedded_nurbs.cps[:, :2],
    )


# Test Extrusion routines
@pytest.mark.parametrize("splinetype", all_2p2d_splines)
def test_create_extrude(splinetype, np_rng, request):
    """
    Test extrusion for different input types and arguments
    """
    # create 2D spline
    spline = request.getfixturevalue(splinetype)

    # Expect Failure - not a spline
    with pytest.raises(NotImplementedError):
        splinepy.helpme.create.extruded([4])

    if splinetype == "bspline_2p2d":
        # Expect Failure - no axis given
        with pytest.raises(ValueError):
            spline.create.extruded()
        # Expect Failure - axis wrong format
        with pytest.raises(ValueError):
            spline.create.extruded(extrusion_vector=[1])

    # Create a random axis
    axis = np_rng.random(3)
    x, y, z = np_rng.random(3)

    # Test results
    assert np.allclose(
        spline.create.extruded(extrusion_vector=axis).evaluate([[x, y, z]]),
        np.hstack((spline.evaluate([[x, y]]), np.zeros((1, 1)))) + z * axis,
    )


# Test Revolution Routine
@pytest.mark.parametrize("splinetype", all_2p2d_splines)
def test_create_revolution(splinetype, np_rng, request):
    """
    Test revolution routines for different input types and arguments
    """
    # Make 2D spline
    spline = request.getfixturevalue(splinetype)

    # Make some lines
    bezier_line = splinepy.Bezier(control_points=[[1, 0], [2, 1]], degrees=[1])
    nurbs_line = bezier_line.nurbs

    # Make a cuboid
    cuboid = splinepy.Bezier(
        control_points=[
            [0, 0, 0],
            [1, 0, 0],
            [0, 1, 0],
            [1, 1, 0],
            [0, 0, 1],
            [1, 0, 1],
            [0, 1, 1],
            [1, 1, 1],
        ],
        degrees=[1, 1, 1],
    )

    # Expect Failure - not a spline
    with pytest.raises(NotImplementedError):
        splinepy.helpme.create.revolved([4])
    # Expect Failure - No rotation axis
    with pytest.raises(ValueError):
        cuboid.create.revolved()

    if splinetype == "bspline_2p2d":
        # Expect Failure - axis wrong format
        with pytest.raises(ValueError):
            spline.create.revolved(axis=[1])
        # Expect Failure - axis too small
        with pytest.raises(ValueError):
            spline.create.revolved(axis=[0, 0, 1e-18])

    # Revolve always around z-axis
    # init rotation matrix
    r_angle = np_rng.random()
    r_center = np.array([1, 0])
    cc, ss = np.cos(r_angle), np.sin(r_angle)
    R = np.array([[cc, -ss, 0], [ss, cc, 0], [0, 0, 1]])
    R2 = np.array([[cc, -ss], [ss, cc]])

    # Test 3D revolutions for bodies
    dim_bumped_cps = np.zeros((spline.control_points.shape[0], 1))
    ref_sol = np.matmul(
        np.hstack((spline.control_points, dim_bumped_cps)), R.T
    )

    revolved_cps = spline.create.revolved(
        axis=[0, 0, 1],
        center=[0, 0, 0],
        angle=r_angle,
        degree=False,
    ).control_points

    assert np.allclose(
        revolved_cps[-(len(ref_sol)) :, :], ref_sol
    ), f"{spline.whatami} failed revolution"

    # Test 2D Revolutions of lines
    for spline_g in (bezier_line, nurbs_line):
        assert np.allclose(
            spline_g.create.revolved(
                angle=r_angle, degree=False
            ).control_points[-2:, :],
            np.matmul(spline_g.control_points, R2.T),
        ), f"{spline_g.whatami} failed revolve"

    # Test 2D Revolutions of lines around center
    for spline_g in (bezier_line, nurbs_line):
        assert np.allclose(
            spline_g.create.revolved(
                angle=r_angle, center=r_center, degree=False
            ).control_points[-2:, :],
            np.matmul(spline_g.control_points - r_center, R2.T) + r_center,
        ), f"{spline_g.whatami} failed revolution around center"


@pytest.mark.parametrize("splinetype", all_splines)
def test_create_parametric_view(splinetype, request):
    """test parametric view"""

    def check_parametric_view(spline, conform):
        p_spl = spline.create.parametric_view(conform=conform)

        # for both conform and pure view
        # spl's pbounds and para's physbound are same
        assert np.allclose(
            p_spl.control_point_bounds, spline.parametric_bounds
        )

        if conform:
            # same spline type
            assert type(p_spl) == type(spline)

            # same degrees
            assert np.allclose(p_spl.ds, spline.ds)

            # same knot_vectors
            if spline.has_knot_vectors:
                for p_kv, kv in zip(p_spl.kvs, spline.kvs):
                    assert np.allclose(p_kv, kv)

            # same weights
            if spline.is_rational:
                assert np.allclose(p_spl.ws, spline.ws)

        else:
            # degrees are one - subtracting 1 should make it all zero
            assert not any(p_spl.ds - 1)

            # same unique knots - implies same p_bounds
            for p_ukv, ukv in zip(p_spl.unique_knots, spline.unique_knots):
                assert np.allclose(p_ukv, ukv)

    spl = request.getfixturevalue(splinetype)
    check_parametric_view(spl, False)
    check_parametric_view(spl, True)


def test_determinant_spline(
    np_rng,
    bezier_2p2d,
    bspline_2p2d,
    bezier_3p3d,
    rational_bezier_3p3d,
    bspline_3p3d,
    nurbs_3p3d,
):
    # Bezier splines
    # arbitrary
    bez_1 = splinepy.Bezier(
        degrees=[2], control_points=[[0, 0], [1.0, 0.5], [1, 0]]
    )
    bez_1 = splinepy.helpme.create.extruded(bez_1, [0, 2])
    bez_1 = splinepy.helpme.create.extruded(bez_1, [0, 0, 3])

    # box
    bez_2 = splinepy.helpme.create.box(3, 3, 3)

    # BSplines
    # C^0 continuous
    bsp_c0 = splinepy.BSpline(
        degrees=[1, 1],
        knot_vectors=[
            [0.0, 0.0, 1.0, 5.0, 5.0],
            [0.0, 0.0, 1.0, 3.0, 3.0],
        ],
        control_points=[
            [0.0, 0.0],
            [0.5, 0.0],
            [1.0, 0.0],
            [0.5, 0.5],
            [1.0, 0.5],
            [1.5, 0.5],
            [0.0, 1.0],
            [0.5, 1.0],
            [1.0, 1.0],
        ],
    )

    # C^1 Continuous
    bsp_c1 = splinepy.BSpline(
        degrees=[3, 1],
        knot_vectors=[
            [0.0, 0.0, 0.0, 0.0, 2.0, 2.0, 4.0, 4.0, 4.0, 4.0],
            [0.0, 0.0, 1.0, 1.0],
        ],
        control_points=[
            [0.0, 0.0],
            [0.5, 0.0],
            [1.0, -1.0],
            [2.0, -3.0],
            [2.5, -3.0],
            [3.0, -3],
            [5.0, 5.0],
            [6.0, 5.0],
            [7.0, 5.0],
            [8.0, 5.0],
            [9.0, 5.0],
            [10.0, 5.0],
        ],
    )
    bsp_c1 = splinepy.helpme.create.extruded(bsp_c1, [0, 0, 3])

    # CPTS Manipulation for almost tangled spline
    bsp_c1.control_points[17] = [3, -3, 1.51]
    bsp_c1.control_points[5] = [3, -3, 1.49]

    bsp_c1_tang = bsp_c1.copy()

    # CPTS Manipulation for slightly tangled spline
    bsp_c1_tang.control_points[5] = [3, -3, 1.51]
    bsp_c1_tang.control_points[17] = [3, -3, 1.49]

    # After degree elevation det(J) < 0??
    bsp_c1.elevate_degrees([0, 1, 1, 2])

    # NURBS
    nurbs_eq_w = splinepy.NURBS(
        degrees=[3, 1],
        knot_vectors=[
            [0.0, 0.0, 0.0, 0.0, 2.0, 2.0, 4.0, 4.0, 4.0, 4.0],
            [0.0, 0.0, 1.0, 1.0],
        ],
        control_points=[
            [0.0, 0.0],
            [0.5, 0.0],
            [1.0, -1.0],
            [2.0, -3.0],
            [2.5, -3.0],
            [3.0, -3],
            [5.0, 5.0],
            [6.0, 5.0],
            [7.0, 5.0],
            [8.0, 5.0],
            [9.0, 5.0],
            [10.0, 5.0],
        ],
        weights=[1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1],
    )

    nurbs_eq_w = splinepy.helpme.create.extruded(nurbs_eq_w, [0, 0, 3])

    # CPTS Manipulation for almost tangled spline
    nurbs_eq_w.control_points[5] = [3, -3, 1.49]
    nurbs_eq_w.control_points[17] = [3, -3, 1.51]

    nurbs_eq_w_tang = nurbs_eq_w.copy()
    # CPTS Manipulation for slightly tangled spline
    nurbs_eq_w_tang.control_points[5] = [3, -3, 1.51]
    nurbs_eq_w_tang.control_points[17] = [3, -3, 1.49]

    # After degree elevation det(J) < 0??
    nurbs_eq_w.elevate_degrees([0, 0, 1, 2])

    # Splines which are not tangled

    for idx, sp_i in enumerate(
        [
            bez_2,
            bsp_c0,
            bsp_c1,
            nurbs_eq_w,
            bezier_2p2d,
            bspline_2p2d,
            bezier_3p3d,
            rational_bezier_3p3d,
            bspline_3p3d,
            nurbs_3p3d,
            bez_1,
            bsp_c1_tang,
            nurbs_eq_w_tang,
        ]
    ):
        det_spl = splinepy.helpme.create.determinant_spline(sp_i)
        rnd_queries = np_rng.random((10, sp_i.dim))

        assert np.allclose(
            det_spl.evaluate(queries=rnd_queries).ravel(),
            np.linalg.det(sp_i.jacobian(queries=rnd_queries)),
        ), f"{sp_i.whatami} at index {idx} failed determinant spline"



import gustaf as _gus
import numpy as _np

from splinepy._base import SplinepyBase as _SplinepyBase
from splinepy.bspline import BSpline as _BSpline
from splinepy.multipatch import Multipatch as _Multipatch
from splinepy.splinepy_core import PySpline as _PySpline
from splinepy.utils.data import cartesian_product as _cartesian_product


class Microstructure(_SplinepyBase):
    """Helper class to facilitatae the construction of microstructures."""

    def __init__(
        self,
        deformation_function=None,
        tiling=None,
        microtile=None,
        parametrization_function=None,
    ):
        """Helper class to facilitatae the construction of microstructures.

        Parameters
        ----------
        deformation_function : spline
          Outer function that describes the contour of the microstructured
          geometry
        tiling : list of integers
          microtiles per parametric dimension
        microtile : spline or list of splines
          Representation of the building block defined in the unit cube
        parametrization_function : Callable (optional)
          Function to describe spline parameters
        """
        if deformation_function is not None:
            self.deformation_function = deformation_function

        if tiling is not None:
            self.tiling = tiling

        if microtile is not None:
            self.microtile = microtile

        if parametrization_function is not None:
            self.parametrization_function = parametrization_function

    @property
    def deformation_function(self):
        """Deformation function defining the outer geometry (contour) of the
        microstructure.

        Parameters
        ----------
        None

        Returns
        -------
        deformation_function : spline
        """
        if hasattr(self, "_deformation_function"):
            return self._deformation_function
        else:
            return None

    @deformation_function.setter
    def deformation_function(self, deformation_function):
        """Deformation function setter defining the outer geometry of the
        microstructure. Must be spline type and as such inherit from
        splinepy.Spline.

        Parameters
        ----------
        deformation_function : spline

        Returns
        -------
        None
        """

        if not isinstance(deformation_function, _PySpline):
            raise ValueError(
                "Deformation function must be splinepy-Spline."
                " e.g. splinepy.NURBS"
            )
        self._deformation_function = deformation_function

        if self._deformation_function.has_knot_vectors:
            self._deformation_function.normalize_knot_vectors()

        self._sanity_check()

    @property
    def tiling(self):
        """Number of microtiles per parametric dimension.

        Parameters
        ----------
        None

        Returns
        -------
        tiling : list<int>
        """
        if hasattr(self, "_tiling"):
            return self._tiling
        else:
            return None

    @tiling.setter
    def tiling(self, tiling):
        """Setter for the tiling attribute, defining the number of microtiles
        per parametric dimension.

        Parameters
        ----------
        tiling : int / list<int>
          Number of tiles for each dimension respectively
        Returns
        -------
        None
        """
        if not isinstance(tiling, list) and not isinstance(tiling, int):
            raise ValueError(
                "Tiling mus be either list of integers of integer " "value"
            )
        self._tiling = tiling
        # Is defaulted to False using function arguments
        self._sanity_check()
        self._logd(f"Successfully set tiling to : {self.tiling}")

    @property
    def microtile(self):
        """Microtile that is either a spline, a list of splines, or a class
        that provides a `create_tile` function."""
        if hasattr(self, "_microtile"):
            return self._microtile
        else:
            self._logi(
                "microtile is empty. "
                "Please checkout splinepy.microstructure.tiles.show() for "
                "predefined tile collections."
            )
            return None

    @microtile.setter
    def microtile(self, microtile):
        """Setter for microtile.

        Microtile must be either a spline, a list of splines, or a class that
        provides (at least) a `create_tile` function and a `dim` member.

        Parameters
        ----------
        microtile : spline / list<splines> / user-object
          arbitrary long list of splines that define the microtile

        Returns
        -------
        None
        """
        # place single tiles into a list to provide common interface
        if isinstance(microtile, (_PySpline, list)):
            microtile = self._make_microtilable(microtile)
        # Assign Microtile object to member variable
        self._microtile = microtile

        self._sanity_check()

    @property
    def parametrization_function(self):
        """Function, that - if required - parametrizes the microtiles.

        In order to use said function, the Microtile needs to provide a couple
        of attributes:

         - evaluation_points - a list of points defined in the unit cube
           that will be evaluated in the parametrization function to provide
           the required set of data points
         - para_dim - dimensionality of the parametrization
           function and number of design variables for said microtile

        Parameters
        ----------
        None

        Returns
        -------
        parametrization_function : Callable
          Function that describes the local tile parameters
        """
        if hasattr(self, "_parametrization_function"):
            return self._parametrization_function
        else:
            return None

    @parametrization_function.setter
    def parametrization_function(self, parametrization_function):
        if not callable(parametrization_function):
            raise TypeError("parametrization_function must be callable")
        self._parametrization_function = parametrization_function
        self._sanity_check()

    @property
    def parameter_sensitivity_function(self):
        """Function, that - if required - determines the parameter sensitivity
        of a set of microtiles

        In order to use said function, the Microtile needs to provide a couple
        of attributes:

         - evaluation_points - a list of points defined in the unit cube
           that will be evaluated in the parametrization function to provide
           the required set of data points
         - para_dim - dimensionality of the parametrization
           function and number of design variables for said microtile
         - parametrization_function - a function that calculates the microtile
           parameters based on the position of the tile within the deformation
           functions parametric space

        Parameters
        ----------
        None

        Returns
        -------
        parameter_sensitivity_function : Callable
          Function that describes the local tile parameters
        """
        if hasattr(self, "_parameter_sensitivity_function"):
            return self._parameter_sensitivity_function
        else:
            return None

    @parameter_sensitivity_function.setter
    def parameter_sensitivity_function(self, parameter_sensitivity_function):
        if parameter_sensitivity_function is None:
            self._parameter_sensitivity_function = None
            self._sanity_check()
            return
        if not callable(parameter_sensitivity_function):
            raise TypeError("parametrization_function must be callable")
        self._parameter_sensitivity_function = parameter_sensitivity_function
        self._sanity_check()

    def _additional_knots(self, knot_span_wise):
        """Determine all required additional knots to be inserted into the
        original spline representation of the deformation function

        Parameters
        ----------
        knot_span_wise: bool
          Decicive, whether tiling is applied to all knot-spans respectively,
          or globally

        Returns
        -------
        additional_knots : list<list>
          returns a list of all required additional knots as a list of list to
          the respective parametric dimension. If no additional knot is to be
          inserted, the list is empty
        """
        additional_knots = []
        # Create Spline that will be used to iterate over parametric space
        ukvs = self.deformation_function.unique_knots
        if knot_span_wise:
            for tt, ukv in zip(self.tiling, ukvs):
                inv_t = 1 / tt
                new_knots = [
                    ukv[i - 1] + j * inv_t * (ukv[i] - ukv[i - 1])
                    for i in range(1, len(ukv))
                    for j in range(1, tt)
                ]
                additional_knots.append(new_knots)
        else:
            self._logd(
                "New knots will be inserted one by one with the objective"
                " to evenly distribute tiles within the parametric domain"
            )
            for i_pd, (tt, ukv) in enumerate(zip(self.tiling, ukvs)):
                n_current_spans = len(ukv) - 1
                if tt == n_current_spans:
                    continue
                elif tt < n_current_spans:
                    self._logw(
                        f"The requested tiling can not be provided, as "
                        f"there are too many knotspans in the deformation"
                        f" function. The tiling in parametric dimension "
                        f"{i_pd} will be set to {n_current_spans}"
                    )
                    self.tiling[i_pd] = n_current_spans
                else:
                    # Determine new knots
                    n_k_span = _np.zeros(n_current_spans, dtype=int)
                    span_measure = _np.diff(ukv)
                    for _ in range(tt - n_current_spans):
                        add_knot = _np.argmax(span_measure)
                        n_k_span[add_knot] += 1
                        span_measure[add_knot] *= n_k_span[add_knot] / (
                            n_k_span[add_knot] + 1
                        )

                    new_knots = []
                    for i, nks in enumerate(n_k_span):
                        new_knots.extend(
                            _np.linspace(ukv[i], ukv[i + 1], nks + 2)[1:-1]
                        )
                    additional_knots.append(new_knots)

        # Return all knots
        return additional_knots

    def _compute_tiling_prerequisites(
        self, knot_span_wise, is_parametrized, macro_sensitivity
    ):
        """
        Prepare all required information to start the construction process. All
        additional properties are returned None, to avoid argument dependent
        return values

        Parameters
        ----------
        knot_span_wise: bool
          Decides if the number of tiles in tiling is supposed to be
          interpreted as the total amount or per knot span
        is_parametrized : bool
          Decides if the microtiles are adapted before insertion into the
          microstructure
        macro_sensitivity :bool
          Provides information required for the calculation of derivatives with
          respect to the control points of the deformation function

        Returns
        -------
        def_fun_patches: list<BezierBasis>
          Extracted Bezier patches, that define the position of the individual
          microtiles
        bezier_matrices: list<numpy/scipy.sparse>
          (None if `macro_sensitivity` is `False`) Matrices that can be used to
          map from the original control points to the control points of the
          extracted bezier patches
        para_space_patches: list<BezierBasis>
          (None if `is_parametrized` is `False`) Linear representation of the
          parametric space that can be used to evaluate the parametrization
          function on the individual elements
        element_resolutions: list<int>
          Number of elements in the respective parametric dimension
        """

        # Prepare the deformation function
        # Transform into a non-uniform splinetype and make sure to work on copy
        if self._deformation_function.is_rational:
            deformation_function_copy = self._deformation_function.nurbs
        else:
            deformation_function_copy = self._deformation_function.bspline

        additional_knots = self._additional_knots(knot_span_wise)

        # First step : insert all required knots into the deformation function
        # spline
        if macro_sensitivity:
            knot_insertion_matrix = (
                deformation_function_copy.knot_insertion_matrix(
                    0, additional_knots[0]
                )
            )
        deformation_function_copy.insert_knots(0, additional_knots[0])
        for i_pd, akv in enumerate(additional_knots[1:], start=1):
            if macro_sensitivity:
                knot_insertion_matrix = (
                    deformation_function_copy.knot_insertion_matrix(i_pd, akv)
                    @ knot_insertion_matrix
                )
            deformation_function_copy.insert_knots(i_pd, akv)

        # Second step (if MS is parametrized)
        if is_parametrized:
            para_bounds = self.deformation_function.parametric_bounds.T
            def_fun_para_space = _BSpline(
                degrees=[1] * self.deformation_function.para_dim,
                knot_vectors=para_bounds.repeat(2, 1),
                control_points=_cartesian_product(para_bounds),
            )
            for i_pd in range(deformation_function_copy.para_dim):
                if len(deformation_function_copy.unique_knots[i_pd][1:-1]) > 0:
                    def_fun_para_space.insert_knots(
                        i_pd,
                        deformation_function_copy.unique_knots[i_pd][1:-1],
                    )

        # Extract the bezier patches
        def_fun_patches = deformation_function_copy.extract.beziers()

        # Precompute the extraction matrices
        if macro_sensitivity:
            bezier_matrices = deformation_function_copy.knot_insertion_matrix(
                beziers=True
            )

            # However, these refer to the macro spline that has knots already
            # inserted into it, so we need to apply the matrix constructed
            # earlier
            for i_matrix in range(len(bezier_matrices)):
                bezier_matrices[i_matrix] = (
                    bezier_matrices[i_matrix] @ knot_insertion_matrix
                )
        else:
            bezier_matrices = None

        # Do the same for the parametric space representation
        if is_parametrized:
            para_space_patches = def_fun_para_space.extract.beziers()
        else:
            para_space_patches = None

        # Determine element resolution
        element_resolutions = [
            len(c) - 1 for c in deformation_function_copy.unique_knots
        ]

        return (
            def_fun_patches,
            bezier_matrices,
            para_space_patches,
            element_resolutions,
        )

    def create(
        self,
        closing_face=None,
        knot_span_wise=None,
        macro_sensitivities=None,
        **kwargs,
    ):
        """Create a Microstructure.

        Parameters
        ----------
        closing_face : string
          If not None, Microtile must provide a function `closing_tile`
          Represents coordinate to be a closed surface {"x", "y", "z"}
        knot_span_wise : bool
          Insertion per knotspan vs. total number per paradim
        macro_sensitivities: bool
          Calculate the derivatives of the structure with respect to the outer
          control point variables
        **kwargs
          will be passed to `create_tile` function

        Returns
        -------
        Microstructure : list<spline>
          finished microstructure based on object requirements
        """
        # Check if all information is gathered
        if not self._sanity_check():
            raise ValueError("Not enough information provided, abort")

        # Set default values
        if knot_span_wise is None:
            knot_span_wise = True
        if macro_sensitivities is None:
            macro_sensitivities = False

        # check if user wants closed structure
        if closing_face is not None:
            if not hasattr(self.microtile, "_closing_tile"):
                raise ValueError(
                    "Microtile does not provide closing tile definition"
                )
            closing_face_dim = {"x": 0, "y": 1, "z": 2}.get(closing_face)
            if closing_face is None:
                raise ValueError(
                    "Invalid format for closing_face argument, (handed: "
                    f"{closing_face}), must be one of"
                    "{'x', 'y', 'z'}"
                )
            if closing_face_dim >= self._deformation_function.para_dim:
                raise ValueError(
                    "closing face must be smaller than the deformation "
                    "function's parametric dimension"
                )
            is_closed = True
        else:
            is_closed = False

        # Check if parametrized
        is_parametrized = self.parametrization_function is not None
        parameter_sensitivities = (
            self.parameter_sensitivity_function is not None
        )

        if parameter_sensitivities and not is_parametrized:
            raise ValueError(
                "Parameter sensitivities can not be interpreted without"
                "parametrization function"
            )

        # Using the additional knots and the bezier extraction operator, we can
        # provide all necessary information on the construction process
        (
            bezier_patches,
            knot_insertion_matrices,
            para_space_patches,
            element_resolutions,
        ) = self._compute_tiling_prerequisites(
            knot_span_wise=knot_span_wise,
            is_parametrized=is_parametrized,
            macro_sensitivity=macro_sensitivities,
        )

        # Use element_resolutions to get global indices of patches
        local_indices = _cartesian_product(
            [_np.arange(er) for er in element_resolutions]
        )

        # Determine the number of geometric derivatives with respect to the
        # design variables
        if parameter_sensitivities:
            n_parameter_sensitivities = self.parameter_sensitivity_function(
                self._microtile.evaluation_points
            ).shape[2]
        else:
            n_parameter_sensitivities = 0

        # Determine the number of geometric derivatives with respect to the
        # deformation function's control points
        if macro_sensitivities:
            n_macro_sensitivities = self.deformation_function.cps.size
        else:
            n_macro_sensitivities = 0

        # Prepare field for derivatives
        if macro_sensitivities or parameter_sensitivities:
            spline_list_derivs = [
                []
                for i in range(
                    n_parameter_sensitivities + n_macro_sensitivities
                )
            ]

        spline_list_ms = []

        # Start actual composition
        for i_patch, def_fun in enumerate(bezier_patches):
            # If the structure is parametrized, provide parameters for the tile
            # construction
            if is_parametrized:
                # Retrieve position from the parametric space patches
                positions = para_space_patches[i_patch].evaluate(
                    self._microtile.evaluation_points
                )
                # Evaluate parametrization function
                tile_parameters = self.parametrization_function(positions)
            else:
                tile_parameters = None

            # If the sensitivities are requested, evaluate the sensitivity
            # function, which must be provided by the user
            if parameter_sensitivities:
                tile_sensitivities = self.parameter_sensitivity_function(
                    positions
                )
                # To avoid unnecessary constructions (if a parameter
                # sensitivity evaluates to zero), perform only those that are
                # in the support of a specific design variable
                support_mask = _np.any(tile_sensitivities != 0.0, axis=(0, 1))
                support = support_mask.nonzero()[0]
                anti_support = (~support_mask).nonzero()[0]

                tile_sens_on_support = tile_sensitivities[:, :, support]
            else:
                tile_sens_on_support = None

            # Check if center or closing tile
            if is_closed:
                # check index
                index = local_indices[i_patch, closing_face_dim]
                face_closure = (
                    closing_face + "_min"
                    if (index == 0)
                    else (
                        closing_face + "_max"
                        if (index + 1) == element_resolutions[closing_face_dim]
                        else None
                    )
                )
            else:
                face_closure = None

            # Determine the microtile (prior to insertion) derivative is set to
            # None if the parameter sensitivity is not requested
            tile, derivatives = self._microtile.create_tile(
                parameters=tile_parameters,
                parameter_sensitivities=tile_sens_on_support,
                closure=face_closure,
                **kwargs,
            )

            # Perform the composition
            if macro_sensitivities:
                basis_function_compositions = []
                for tile_patch in tile:
                    composed, basis_function_composition = def_fun.compose(
                        tile_patch, compute_sensitivities=True
                    )
                    spline_list_ms.append(composed)
                    basis_function_compositions.append(
                        basis_function_composition
                    )
            else:
                for tile_patch in tile:
                    spline_list_ms.append(def_fun.compose(tile_patch))

            # Determine the geometric sensitivities using the chain rule
            if parameter_sensitivities:
                n_splines_per_tile = len(tile)
                empty_splines = [None] * n_splines_per_tile
                for j in anti_support:
                    spline_list_derivs[j].extend(empty_splines)
                for j, deris in enumerate(derivatives):
                    for tile_v, tile_deriv in zip(tile, deris):
                        spline_list_derivs[support[j]].append(
                            def_fun.composition_derivative(tile_v, tile_deriv)
                        )

            # Compute the derivatives with respect to the deformation
            # function's ctps
            if macro_sensitivities:
                for patch_info in basis_function_compositions:
                    # patch_info is a list of splines all representing the
                    # derivatives of the composed spline with respect to a
                    # component of a control point within the bezier patch
                    # of the deformation function.
                    cps = _np.hstack([p.cps for p in patch_info])
                    # We use the matrices to map the contributions of the
                    # bezier ctps to the deformation functions
                    mapped_cps = cps @ knot_insertion_matrices[i_patch]

                    # Last step: create beziers to be added to the derivative
                    # fields
                    for j_cc in range(n_macro_sensitivities):
                        control_points = _np.zeros(
                            (cps.shape[0], self._deformation_function.dim)
                        )
                        ii_ctps, jj_dim = divmod(
                            j_cc, self._deformation_function.dim
                        )
                        control_points[:, jj_dim] = mapped_cps[:, ii_ctps]
                        spline_list_derivs[
                            n_parameter_sensitivities + j_cc
                        ].append(
                            type(patch_info[0])(
                                patch_info[0].degrees, control_points
                            )
                        )

        # Use a multipatch object to bundle all information
        self._microstructure = _Multipatch(splines=spline_list_ms)

        # Add fields if requested
        if macro_sensitivities or parameter_sensitivities:
            self._microstructure.add_fields(
                spline_list_derivs, self.deformation_function.dim
            )

        return self._microstructure

    def show(self, use_saved=False, **kwargs):
        """
        Shows microstructure. Consists of deformation_function, microtile, and
        microstructure. Supported only by vedo.

        Parameters
        ----------
        use_saved: bool
        **kwargs: kwargs
          Will be passed to show function

        Returns
        -------
        plt: vedo.Plotter
        """
        if use_saved:
            if hasattr(self, "_microstructure"):
                microstructure = self._microstructure
            else:
                raise ValueError("No previous microstructure saved")
        else:
            # Create on the fly
            microstructure = self.create(**kwargs)

        # Precompute splines
        microtile = self.microtile.create_tile(**kwargs)
        deformation_function = self.deformation_function

        # turn off control points for clarity
        microstructure.show_options["control_points"] = False

        # Show in vedo
        return _gus.show(
            ["Deformation Function", deformation_function],
            ["Microtile", microtile[0]],
            ["Composed Microstructure", microstructure],
            **kwargs,
        )

    def _sanity_check(self):
        """Check all members and consistency of user data.

        Parameters
        ----------
        updated_properties : bool
          Sets the updated_properties variable to value, which indicates,
          wheither the microstructure needs to be rebuilt

        Returns
        -------
        passes: bool
        """
        if (
            (self.deformation_function is None)
            or (getattr(self, "_microtile", None) is None)
            or (self.tiling is None)
        ):
            self._logd(
                "Current information not sufficient,"
                " awaiting further assignments"
            )
            return False
        # Check if microtile object fulfils requirements
        if not hasattr(self._microtile, "create_tile"):
            raise ValueError(
                "Microtile class does not provide the necessary "
                "attribute `create_tile`, that is required for "
                "microstructure construction"
            )
        if not hasattr(self._microtile, "dim"):
            raise ValueError(
                "Microtile class does not provide the necessary "
                "attribute `dim`, defining the dimensionality of "
                "the created tile"
            )

        # Check if parametric dimensions are consistent
        if not self.deformation_function.para_dim == self._microtile.dim:
            raise ValueError(
                "Microtile dimension must match parametric dimension of "
                "deformation function to enable composition"
            )

        # Check if tiling is consistent
        if isinstance(self.tiling, int):
            self.tiling = [self.tiling] * self.deformation_function.para_dim
        if len(self.tiling) != self.deformation_function.para_dim:
            raise ValueError(
                "Tiling list must have one entry per parametric dimension"
                " of the deformation function"
            )

        if self.parametrization_function is None:
            # All checks have passed
            return True

        self._logd("Checking compatibility of parametrization function")
        if not hasattr(self._microtile, "evaluation_points"):
            raise ValueError(
                "Microtile class does not provide the necessary "
                "attribute `evaluation_points`, that is required for"
                " a parametrized microstructure construction"
            )
        result = self._parametrization_function(
            self._microtile.evaluation_points
        )
        if not isinstance(result, _np.ndarray):
            raise ValueError(
                "Function outline of parametrization function must be "
                "`f(np.ndarray)->np.ndarray`"
            )
        if self.parameter_sensitivity_function is None:
            # All checks have passed
            return True

        # Check sensitivity function
        result = self.parameter_sensitivity_function(
            self._microtile.evaluation_points
        )
        if (not isinstance(result, _np.ndarray)) or (not result.ndim == 3):
            raise ValueError(
                "Function outline of parameter sensitivity function must "
                "be `f(np.ndarray)->np.ndarray`, with dimension 3 and shape:"
                "n_evaluation_points x n_info_per_evaluation_point x "
                "n_sensitivities"
            )

        return True

    def _make_microtilable(self, microtile):
        """Creates a Microtile object on the fly if user only provides (a list
        of) splines. Internal use only.

        Parameters
        ----------
        microtile : spline / list<splines>
          Microtile definition of a spline
        """
        return _UserTile(microtile)


class _UserTile(_SplinepyBase):
    def __init__(self, microtile):
        """
        On the fly created class of a user tile
        Parameters
        ----------
        microtile : spline , list<splines>
        """
        # Assign microtiles
        self._user_tile = []

        if not isinstance(microtile, list):
            microtile = [microtile]

        for m in microtile:
            if not isinstance(m, _PySpline):
                raise ValueError(
                    "Microtiles must be (list of) "
                    "splinepy-Splines. e.g. splinepy.NURBS"
                )
            # Extract beziers for every non Bezier patch else this just
            # returns itself
            self._user_tile.extend(m.extract.beziers())
        self._dim = microtile[0].dim
        for m in microtile:
            if m.dim != self._dim:
                raise ValueError("Dimensions of spline lists inconsistent")

    @property
    def dim(self):
        return self._dim

    def create_tile(self, **kwargs):  # noqa ARG002
        """Create a tile on the fly.
        Derivative of a constant tile can not be requested, so derivative must
        be None
        Parameters
        ----------
        None

        Returns
        -------
        Microtile: list<Bezier>
        Derivative : None
        """
        for k, v in kwargs.items():
            self._logd(f"Additional argument {k} : {v} will be ignored")
        return (self._user_tile.copy(), None)



"""
Isophotes are a tool to visualize the continuity of Surfaces. They represent
areas of equal brightness, supposing a parallel light source. This can be used
to check interface connections between two patches. If isophotes are at least
G0 continuous, the connection is G1 continuous, further, if they are G1 smooth,
the connection is G2 continuous. This visual aid can help find kinks in
surfaces.

Further, this example shows how to compute the Mean and Gaussian Curvature of
a given surface based on the description in
    "Rogers, David F. An introduction to NURBS: with historical perspective.
     Morgan Kaufmann, 2001"
"""

import numpy as np

import splinepy as sp

np_rng = np.random.default_rng()
# Create a random spline surface in 3D
a = sp.helpme.create.box(1, 1).bspline
a.elevate_degrees([0, 1, 0, 1])
a.insert_knots(0, [0.33333, 0.66667])
a.insert_knots(1, [0.33333, 0.66667])
a.control_points = np.hstack([a.cps, np_rng.random((a.cps.shape[0], 1)) * 0.5])


# Function to create a Callable for the SplineDataAdaptor
def isophotes_to_direction(normal=None, interval=0.02):
    if normal is None:
        normal = [1, 0, 0]
    normal = normal / np.linalg.norm(normal)

    def isophotes_scalars(data, on):
        v1 = data.derivative(on, [1, 0])
        v2 = data.derivative(on, [0, 1])
        normals = np.cross(v1, v2)
        test = np.arccos(
            np.matmul(normals, normal) / np.linalg.norm(normals, axis=1)
        )
        return np.floor(test / interval) % 2

    return isophotes_scalars


# Gaussian Curvature
def gaussian_curvature(data, on):
    v1 = data.derivative(on, [1, 0])
    v2 = data.derivative(on, [0, 1])
    v3 = data.derivative(on, [2, 0])
    v4 = data.derivative(on, [0, 2])
    v5 = data.derivative(on, [1, 1])
    # Using notation from "An introduction to NURBS historical perspective"
    normals = np.cross(v1, v2)
    A = np.einsum("ij,ij->i", normals, v3)
    B = np.einsum("ij,ij->i", normals, v4)
    C = np.einsum("ij,ij->i", normals, v5)

    k_g = (A * C - B**2) / np.linalg.norm(normals, axis=1) ** 4
    return k_g


# Mean Curvature
def mean_curvature(data, on=None):
    v1 = data.derivative(on, [1, 0])
    v2 = data.derivative(on, [0, 1])
    v3 = data.derivative(on, [2, 0])
    v4 = data.derivative(on, [0, 2])
    v5 = data.derivative(on, [1, 1])
    # Using notation from "An introduction to NURBS historical perspective"
    normals = np.cross(v1, v2)
    A = np.einsum("ij,ij->i", normals, v3)
    B = np.einsum("ij,ij->i", normals, v4)
    C = np.einsum("ij,ij->i", normals, v5)
    v1v1 = np.einsum("ij,ij->i", v1, v1)
    v1v2 = np.einsum("ij,ij->i", v1, v2)
    v2v2 = np.einsum("ij,ij->i", v2, v2)

    k_g = (A * v1v1 - 2 * v1v2 * B + C * v2v2) / (
        2 * np.linalg.norm(normals, axis=1) ** 3
    )
    return k_g


# Spline Adaptor to plot function
isophotes = sp.SplineDataAdaptor(
    a, function=isophotes_to_direction(normal=[0, 0, 1])
)
mean_curvature = sp.SplineDataAdaptor(a, function=mean_curvature)
gaussian_curvature = sp.SplineDataAdaptor(a, function=gaussian_curvature)

# Set Plot options
a.spline_data["isophotes"] = isophotes
a.spline_data["gaussian_curvature"] = gaussian_curvature
a.spline_data["mean_curvature"] = mean_curvature
a.show_options["lighting"] = "off"
a.show_options["control_points"] = False
a.show_options["knots"] = False
a.show_options["cmap_n_colors"] = 2
a.show_options["cmap"] = "turbo"

# Show Isophotes
a.show_options["data"] = "isophotes"
a.show(resolutions=1000, title="Isophotes")

# Show Gaussian Curvature
a.show_options["cmap_n_colors"] = 20
a.show_options["data"] = "gaussian_curvature"
a.show_options["scalarbar"] = True
a.show(resolutions=100, title="Gaussian Curvature")

# Show Curvature
a.show_options["data"] = "mean_curvature"
a.show_options["scalarbar"] = True
a.show(resolutions=100, title="Mean Curvature")
